#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
# Copyright (C) 2015 su_v, suv-sf@users.sf.net
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import inkex
from inkex import turtle as pturtle


class TurtleRtree(inkex.GenerateExtension):
    """Create RTree Turtle path"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--size", type=float, default=100.0, help="initial branch size"
        )
        pars.add_argument(
            "--minimum", type=float, default=40.0, help="minimum branch size"
        )
        pars.add_argument(
            "--pentoggle",
            type=inkex.Boolean,
            default=False,
            help="Lift pen for backward steps",
        )

    def generate(self):
        self.options.size = self.svg.unittouu(str(self.options.size) + "px")
        self.options.minimum = self.svg.unittouu(str(self.options.minimum) + "px")
        point = self.svg.namedview.center

        style = inkex.Style(
            {
                "stroke-linejoin": "miter",
                "stroke-width": str(self.svg.unittouu("1px")),
                "stroke-opacity": "1.0",
                "fill-opacity": "1.0",
                "stroke": "#000000",
                "stroke-linecap": "butt",
                "fill": "none",
            }
        )
        tur = pturtle.pTurtle()
        tur.pu()
        tur.setpos(point)
        tur.pd()
        tur.rtree(self.options.size, self.options.minimum, self.options.pentoggle)
        return inkex.PathElement(d=tur.getPath(), style=str(style))


if __name__ == "__main__":
    TurtleRtree().run()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>

# SPDX-License-Identifier: GPL-2.0-or-later


from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from ..parser.types import AFObject
from .curve import AFCurve
from .fill import AFColor, AFGradient, parse_fdsc
from .shape import AFRectangle, AFShape, parse_shape
from .styles import AFStroke
from .util import bbox_helper


class AFTextStretch(Enum):
    ULTRA_CONDENSED = 1
    EXTRA_CONDENSED = 2
    CONDENSED = 3
    SEMI_CONDENSED = 4
    NORMAL = 5
    SEMI_EXPANDED = 6
    EXPANDED = 7
    EXTRA_EXPANDED = 8
    ULTRA_EXPANDED = 9


@dataclass
class AFFont:
    family: str
    italic: bool
    pano: list[int]
    post: str  # post: "family-NarroBoldItalic"
    weight: int
    stretch: AFTextStretch

    @classmethod
    def from_af(cls, font: AFObject) -> AFFont:
        assert font.get_type() == "Font"
        return cls(
            post=font["Post"],
            family=font["Famy"],
            weight=font["Wegt"],
            italic=font["Ital"],
            stretch=AFTextStretch(font["Widh"]),
            pano=font["Pano"]["byte"],
        )


class AFParaAlign(Enum):
    LEFT = 0
    CENTER = 1
    RIGHT = 2
    JUSTIFY_LEFT = 3
    JUSTIFY_CENTER = 4
    JUSTIFY_RIGHT = 5
    JUSTIFY_ALL = 6
    TOWARDS_SPLINE = 7
    AWAY_FROM_SPLINE = 8


class AFTextVAlign(Enum):
    TOP = 0
    CENTER = 1
    BOTTOM = 2
    JUSTIFY = 3


class AFTextLineStyle(Enum):
    NONE = 0
    SINGLE = 1
    DOUBLE = 2


class AFParaLeadingType(Enum):
    MULTIPLE = 0
    PERCENT_HEIGHT = 1
    EXACTLY = 2
    AT_LEAST = 3


@dataclass
class AFParaLeading:
    type: AFParaLeadingType
    value: float

    @classmethod
    def from_af(
        cls, type: AFParaLeadingType, multiple: float, exact: float
    ) -> AFParaLeading:
        return cls(
            type=type,
            value=multiple
            if type in [AFParaLeadingType.PERCENT_HEIGHT, AFParaLeadingType.MULTIPLE]
            else exact
            if type in [AFParaLeadingType.EXACTLY, AFParaLeadingType.AT_LEAST]
            else 0.0,
        )

    def get_line_height(self, font_size: float) -> float:
        if self.type in {AFParaLeadingType.PERCENT_HEIGHT, AFParaLeadingType.MULTIPLE}:
            return self.value * font_size
        if self.type == AFParaLeadingType.AT_LEAST:
            return max(self.value, font_size)
        return self.value


@dataclass
class AFParaAtt:
    index: int

    para_leading: AFParaLeading
    align: AFParaAlign

    @classmethod
    def from_af(cls, paar: AFObject) -> AFParaAtt:
        indx = paar["Indx"]
        item = paar["Item"]
        return AFParaAtt(
            index=indx,
            para_leading=AFParaLeading.from_af(
                AFParaLeadingType(item["Ints"][1]), item["Doub"][0], item["Doub"][1]
            ),
            align=AFParaAlign(item["Ints"][0]),
        )


@dataclass
class AFParaAtts:
    runs: List[AFParaAtt]

    @classmethod
    def from_af(cls, glas: AFObject) -> AFParaAtts:
        runs = []
        for run in glas["Runs"]:
            runs.append(AFParaAtt.from_af(run))
        return cls(runs)


@dataclass
class AFGlyphAtt:
    index: int
    font: AFFont
    rfnt: AFFont  # Unknown

    font_size: float
    fill: Optional[AFColor | AFGradient]
    bg_fill: Optional[AFColor | AFGradient]
    stroke_fill: Optional[AFColor | AFGradient]
    stroke: AFStroke

    underline: AFTextLineStyle
    underline_fill: Optional[AFColor | AFGradient]

    line_through: AFTextLineStyle
    line_through_fill: Optional[AFColor | AFGradient]

    lang: str

    @classmethod
    def from_af(cls, glar: AFObject) -> AFGlyphAtt:
        indx = glar["Indx"]
        item = glar["Item"]
        return cls(
            index=indx,
            font=AFFont.from_af(item["DFnt"]),
            rfnt=AFFont.from_af(item["RFnt"]),
            font_size=item["Doub"][0],
            fill=parse_fdsc(item["Objs"][0]),
            bg_fill=parse_fdsc(item["Objs"][4]),
            stroke_fill=parse_fdsc(item["Objs"][1]),
            stroke=AFStroke.from_af(item["Objs"][2]),
            underline=AFTextLineStyle(item["Ints"][0]),
            line_through=AFTextLineStyle(item["Ints"][1]),
            underline_fill=parse_fdsc(item["Objs"][5]),
            line_through_fill=parse_fdsc(item["Objs"][6]),
            lang=item["Stri"][1],
            # optical_alignment=item["Objs"][8],
        )


@dataclass
class AFGlyphAtts:
    runs: List[AFGlyphAtt]

    @classmethod
    def from_af(cls, glas: AFObject) -> AFGlyphAtts:
        runs = []
        for run in glas["Runs"]:
            runs.append(AFGlyphAtt.from_af(run))
        return cls(runs)


class AFTextType(Enum):
    ARTISTIC = "TxtA"
    FRAMED = "TxtF"
    FOLLOW_CURVE = "CPTx"
    FOLLOW_CURVE2 = "SPTN"
    IN_SHAPE = "TxtS"
    IN_CURVE = "TxtC"


TEXT_TYPES = set(t.value for t in AFTextType)


@dataclass
class AFText:
    type: AFTextType

    glyphs: List[str]
    glyph_atts: List[AFGlyphAtts]
    para_atts: List[AFParaAtts]

    text_path: Optional[AFTextPath]
    shape_inside: AFShape | AFCurve

    # TODO: Unused fields
    colw: Optional[float] = None
    gutw: Optional[float] = None
    valign: AFTextVAlign = AFTextVAlign.TOP

    @classmethod
    def from_af(cls, child: AFObject) -> AFText:
        text_type = AFTextType(child.get_type())
        txth = child.get("TxtH")
        assert txth is not None or text_type == AFTextType.ARTISTIC, "Text info missing"

        b_box = (
            (txth.get("FrmB") if txth else None)
            or child.get("ShpB")
            or child.get("CvsB")
            or child.get("bbox")
        )
        assert b_box is not None, "Text bbox missing"

        shape_inside: AFShape | AFCurve
        if text_type == AFTextType.IN_SHAPE:
            shape_inside = parse_shape(child["Shpe"], child["ShpB"])
        elif text_type == AFTextType.IN_CURVE:
            shape_inside = AFCurve.from_af(child["Crvs"])
        else:
            shape_inside = AFRectangle(bbox_helper(b_box))

        text_path = (
            AFTextPath.from_af(txth)
            if text_type in [AFTextType.FOLLOW_CURVE, AFTextType.FOLLOW_CURVE2]
            else None
        )

        blocks = child["StSt"]["Blok"]
        return cls(
            type=text_type,
            glyphs=[block["Glyp"]["Utf8"] for block in blocks],
            glyph_atts=[AFGlyphAtts.from_af(block["GAtt"]) for block in blocks],
            para_atts=[AFParaAtts.from_af(block["PAtt"]) for block in blocks],
            text_path=text_path,
            shape_inside=shape_inside,
            colw=txth.get("ColW") if txth else None,
            gutw=txth.get("GutW") if txth else None,
            valign=AFTextVAlign(txth["Algn"].id)
            if txth and "Algn" in txth
            else AFTextVAlign.TOP,
        )


@dataclass
class AFTextPath:
    path: AFCurve
    starts: List[float]
    ends: List[float]
    reverses: List[bool]

    @classmethod
    def from_af(cls, txth: AFObject) -> AFTextPath:
        return cls(
            path=AFCurve.from_af(txth["PCrv"]),
            starts=txth["Strt"],
            ends=txth["Ends"],
            reverses=txth["Revr"],
        )



#
# Copyright (C) 2019 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Target a directory to install resources into.
"""

import os
import sys
import json
import logging
from shutil import which

from inkex.command import call

from .archive import Archive, UnrecognizedArchiveFormat
from .remote import RemoteArchive, LocalFile
from .utils import INKSCAPE_PROFILE, CACHE_DIR, ExtensionInx
from .package import PackageItem, OrphanedItem, PythonItem, PythonPackage


class BasicTarget(object):
    """
    A location where to install something, plus how to search for it.
    """

    version_specific = False

    def __init__(self, category, label, path, is_search=False, filters=()):
        self.category = category
        self.label = label
        self.path = os.path.join(INKSCAPE_PROFILE, path)
        self.is_search = is_search
        self.archive = RemoteArchive(category)
        self.filters = filters
        self._installed = None

    def search(self, query, filtered=False):
        """Search the online archive for things"""
        for pkg in self.archive.search(query, filtered):
            pkg.set_installer(self._install)
            yield pkg

    def is_writable(self):
        """Can the folder be written to (for installation)"""
        try:
            return os.access(self.path, os.W_OK)
        except IOError:
            return False

    def _install(self, filename, info):
        if not info.get("id"):
            info["id"] = self.generate_id(filename)

        if not info.get("id"):
            raise ValueError("Id is a required field for packages")

        if filename.endswith(".zip"):
            location = info["id"]
            fname = "info.json"
            info["files"] = list(self.install_zip_files(filename, location))
        else:
            location = "pkg"
            fname = info["id"] + ".json"
            info["files"] = [
                self.write_file(filename, os.path.basename(filename))
            ]

        self.write_file(json.dumps(info).encode("utf8"), fname, extra=location)

        return "Package installed! Remember to restart inkscape to use it!"

    def _uninstall(self, info, json_file):
        self._installed = None
        for fname in info.get("files"):
            self.remove_file(fname)
        if json_file and os.path.isfile(json_file):
            self.remove_file(json_file)
        return True

    def install_zip_files(self, filename, location):
        """Install the files in the zip filename as non-pip files"""
        with Archive(filename) as archive:
            for filename in archive.filenames():
                yield self.write_file(
                    archive.read(filename), filename, extra=location
                )

    def generate_id(self, filename):
        """User submitted zip file, generate an id as best we can"""
        return filename.replace(".zip", "")

    def list_installed(self, cached=True):
        """
        Loops through all the files in the target path and finds all the installed items.
        """
        if cached and self._installed:
            yield from self._installed
            return

        self._installed = []
        for item in self._list_installed():
            self._installed.append(item)
            yield item

    def _list_installed(self):
        orphans = OrphanedItem(self.path)
        for root, subs, files in os.walk(self.path):
            for fname in files:
                fpath = os.path.join(root, fname)
                name = self.unprefix_path(fpath)
                if fname.endswith(".json"):
                    if os.path.basename(fpath) == "package.json":
                        continue
                    with open(fpath, "rb") as fhl:
                        data = fhl.read()
                        info = json.loads(data)
                        if not PackageItem.is_pkg(info):
                            continue

                        item = PackageItem(
                            info, remote=self._remote_or_local_file(root)
                        )
                        item.set_uninstaller(self._uninstall, fpath)
                        yield item

                        for pkg_file in item.get_files(missing=True):
                            orphans.remove_file(pkg_file, item)
                            orphans.remove_file(
                                os.path.join(root, pkg_file), item
                            )
                else:
                    orphans.add_file(name)

        for fname, item in orphans.get_missing():
            if item is not None:
                item._missing.append(fname)

        if orphans.get_files(filters=self.filters):
            yield orphans

    def _remote_or_local_file(self, basedir):
        # If a json file specifies something that's local, it's "ALWAYS" local to the
        # json file as a basedir.
        def _inner(url):
            if "://" not in url:
                return LocalFile(basedir, url)
            return self.archive._remote_file(url)

        return _inner

    def write_file(self, source, source_name=None, extra=None):
        target = os.path.join(self.path, extra) if extra else self.path

        if isinstance(source, str):
            if not source_name:
                source_name = source
            with open(source, "rb") as fhl:
                source = fhl.read()

        path = os.path.join(target, source_name)
        filedir = os.path.dirname(path)
        if not os.path.isdir(filedir):
            os.makedirs(filedir)

        # Ignore paths
        if not os.path.isdir(path):
            with open(path, "wb") as whl:
                whl.write(source)

        return self.unprefix_path(path)

    def remove_file(self, filename):
        """
        Remove the given file and clean up
        """
        if not filename.startswith(self.path):
            filename = os.path.join(self.path, filename)
        if os.path.isfile(filename):
            os.unlink(filename)

        # Recursively clean up directories (if empty)
        path = os.path.dirname(filename)
        while path.lstrip("/") != self.path.lstrip("/"):
            if os.listdir(path):
                break
            os.rmdir(path)
            path = os.path.dirname(path)

    def unprefix_path(self, path):
        """
        Removes the prefix of the given path, if it's based in self.path
        """
        # Strip just the OS seperator, but what if the files were moved from another OS?
        return path.replace(self.path, "").lstrip("/").lstrip("\\")


class ExtensionsTarget(BasicTarget):
    """
    Extra functional target for extensions (pip based)
    """

    version_specific = True

    def get_pip(self):
        path = os.path.abspath(os.path.join(self.path, "bin"))
        return which("pip", path=path + ":" + os.environ["PATH"])

    def _install(self, filename, info):
        if self.is_pip_package(filename):
            results = self.pip_install(filename)
            if results:
                info["pip"] = True
                info["id"] = results.strip().split()[-1]
                fname = info["id"] + ".json"
                self.write_file(
                    json.dumps(info).encode("utf8"), fname, extra="lib"
                )
                return (
                    "Python Package installed! "
                    "Remember to restart inkscape to use it!"
                )
            return "Failed to install, something is wrong with your setup."

        return super()._install(filename, info)

    def _uninstall(self, info, json_file):
        self._installed = None
        if not info.get("pip"):
            return super()._uninstall(info, json_file)
        self.pip_uninstall(info["name"])
        if json_file and os.path.isfile(json_file):
            self.remove_file(json_file)
        return True

    def is_pip_package(self, filename):
        """Return true if this is a detectable pip package"""
        if filename.endswith(".whl"):
            return True
        try:
            with Archive(filename) as archive:
                for filename in archive.filenames():
                    if filename.endswith("setup.py"):
                        return True
        except UnrecognizedArchiveFormat:
            return False
        return False

    def pip_install(self, filename):
        """Install the filename as a pip package"""
        pip = self.get_pip()
        if pip is None:
            logging.error(
                "This package requires python VirtualEnv which is not available on your system."
            )
            return None
        results = call(
            pip,
            "install",
            ("isolated", True),
            ("disable-pip-version-check", True),
            ("cache-dir", CACHE_DIR),
            filename,
        ).decode("utf8")
        return results

    def pip_uninstall(self, name):
        """Uninstall the given pip package name"""
        results = call(
            self.get_pip(),
            "uninstall",
            ("disable-pip-version-check", True),
            name,
        ).decode("utf8")
        return results

    def generate_id(self, filename):
        """Extensions have an id internally, try and use it"""
        try:
            with Archive(filename) as archive:
                inxes = [
                    item
                    for item in archive.filenames()
                    if item.endswith(".inx")
                ]
                if not inxes:
                    raise IOError(
                        "Refusing to install extension without inx file!"
                    )
                inx = ExtensionInx(archive.read(inxes[0]).decode("utf-8"))
            return inx.ident
        except UnrecognizedArchiveFormat:
            raise IOError(
                "Refusing the install extension without inx file (unknown archive)"
            )
        except:  # noqa: E722
            raise IOError("Refusing the install extension with bad inx file!")

    def _list_installed(self):
        """
        Add pip packages to file lists.
        """
        orphans = None
        packages = {}
        all_deps = set()
        all_files = set()

        # First collect a list of python packages installed
        for node in self.get_python_paths():
            if node.endswith(".dist-info") or node.endswith(".egg-info"):
                package = PythonPackage(node, self.path)
                packages[package.name] = package
                all_files |= set(package.package_files())
                for dep, _ in package.get_depedencies():
                    all_deps.add(dep)

        # Now return all non pip packaged extensions (from super)
        for item in super()._list_installed():
            if item.info.get("pip", False):
                if self.info.ident not in packages:
                    print(f"Can't find python package: {item.ient}")
                    continue

                pip_pkg = packages[item.ident]
                item.info["version"] = pip_pkg.version

            if isinstance(item, OrphanedItem):
                orphans = item
            else:
                yield item

        # Remove all orphaned files that were installed by pip packages
        if orphans is not None:
            for fn in all_files:
                orphans.remove_file(fn)
            if orphans.get_files(filters=self.filters):
                # Yield if we still have orphans
                yield orphans

        # Now what to do with all these remaining packages, pretend their installed?
        for name, package in packages.items():
            for inx in package.get_inx():
                item = PythonItem(package)
                item.set_uninstaller(self._uninstall, None)
                yield item
                break

        for dep in all_deps:
            if dep not in packages:
                # These packages are often just installed into the system, nothing to say.
                # XXX But, there is a future where pip could be interigated.
                # logging.error(f"Missing python depedency: {dep}")
                continue
            packages.pop(dep)

    def get_python_paths(self):
        """Returns paths related to the python packages"""
        pyver = "python" + sys.version[:3]
        for varient in [
            os.path.join(self.path, "lib", pyver, "site-packages"),
        ]:
            if os.path.isdir(varient):
                for subpath in os.listdir(varient):
                    yield os.path.join(varient, subpath)

    def get_package(self, name, version=None):
        """Test every package in this list if it matches the name and version"""
        for package in self.iter():
            found = package.is_package(name, version=version)
            if found:
                return package
        return None



#!/usr/bin/env python3
#
# Copyright (C) 2010 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Create webslicer group"""

import inkex
from inkex.localization import inkex_gettext as _
from webslicer_effect import WebSlicerMixin


class CreateGroup(WebSlicerMixin, inkex.EffectExtension):
    """Create new webslicer group"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--html-id", dest="html_id")
        pars.add_argument("--html-class", dest="html_class")
        pars.add_argument("--width-unity", dest="width_unity")
        pars.add_argument("--height-unity", dest="height_unity")
        pars.add_argument("--bg-color", dest="bg_color")

    def effect(self):
        if not self.svg.selection:
            raise inkex.AbortExtension(
                _(
                    'You must to select some "Slicer rectangles" '
                    'or other "Layout groups".'
                )
            )

        base_elements = self.get_slicer_layer().descendants()
        for key, node in self.svg.selection.id_dict().items():
            if node not in base_elements:
                raise inkex.AbortExtension(
                    _(f'The element "{key}" is not in the Web Slicer layer')
                )
            g_parent = node.getparent()

        group = g_parent.add(inkex.Group())
        desc = group.add(inkex.Desc())
        desc.text = self.get_conf_text_from_list(
            ["html_id", "html_class", "width_unity", "height_unity", "bg_color"]
        )

        for node in self.svg.selection.values():
            group.insert(1, node)


if __name__ == "__main__":
    CreateGroup().run()



# coding=utf-8
#
# Copyright (C) 2010 Nick Drobchenko, nick@cnc-club.ru
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Basic common utility functions for calculated things
"""

from collections import OrderedDict, defaultdict
import os
import sys
import random
import re
import math

from argparse import ArgumentTypeError
from itertools import tee, cycle

import numpy as np

ABORT_STATUS = -5

(X, Y) = range(2)
PY3 = sys.version_info[0] == 3

# pylint: disable=line-too-long
# Taken from https://www.w3.org/Graphics/SVG/1.1/paths.html#PathDataBNF
DIGIT_REX_PART = r"[0-9]"
DIGIT_SEQUENCE_REX_PART = rf"(?:{DIGIT_REX_PART}+)"
INTEGER_CONSTANT_REX_PART = DIGIT_SEQUENCE_REX_PART
SIGN_REX_PART = r"[+-]"
EXPONENT_REX_PART = rf"(?:[eE]{SIGN_REX_PART}?{DIGIT_SEQUENCE_REX_PART})"
FRACTIONAL_CONSTANT_REX_PART = rf"(?:{DIGIT_SEQUENCE_REX_PART}?\.{DIGIT_SEQUENCE_REX_PART}|{DIGIT_SEQUENCE_REX_PART}\.)"
FLOATING_POINT_CONSTANT_REX_PART = rf"(?:{FRACTIONAL_CONSTANT_REX_PART}{EXPONENT_REX_PART}?|{DIGIT_SEQUENCE_REX_PART}{EXPONENT_REX_PART})"
NUMBER_REX = re.compile(
    rf"(?:{SIGN_REX_PART}?{FLOATING_POINT_CONSTANT_REX_PART}|{SIGN_REX_PART}?{INTEGER_CONSTANT_REX_PART})"
)
# pylint: enable=line-too-long


def _pythonpath():
    for pth in os.environ.get("PYTHONPATH", "").split(":"):
        if os.path.isdir(pth):
            yield pth


def get_user_directory():
    """Return the user directory where extensions are stored.

    .. versionadded:: 1.1"""
    if "INKSCAPE_PROFILE_DIR" in os.environ:
        return os.path.abspath(
            os.path.expanduser(
                os.path.join(os.environ["INKSCAPE_PROFILE_DIR"], "extensions")
            )
        )

    home = os.path.expanduser("~")
    for pth in _pythonpath():
        if pth.startswith(home):
            return pth
    return None


def get_inkscape_directory():
    """Return the system directory where inkscape's core is.

    .. versionadded:: 1.1"""
    for pth in _pythonpath():
        if os.path.isdir(os.path.join(pth, "inkex")):
            return pth
    raise ValueError("Unable to determine the location of Inkscape")


class KeyDict(dict):
    """
    A normal dictionary, except asking for anything not in the dictionary
    always returns the key itself. This is used for translation dictionaries.
    """

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return key


def parse_percent(val: str):
    """Parse strings that are either values (i.e., '3.14159') or percentages
    (i.e. '75%') to a float.

    .. versionadded:: 1.2"""
    val = val.strip()
    if val.endswith("%"):
        return float(val[:-1]) / 100
    return float(val)


def Boolean(value):
    """ArgParser function to turn a boolean string into a python boolean"""
    if value.upper() == "TRUE":
        return True
    if value.upper() == "FALSE":
        return False
    return None


def to_bytes(content):
    """Ensures the content is bytes

    .. versionadded:: 1.1"""
    if isinstance(content, bytes):
        return content
    return str(content).encode("utf8")


def debug(what):
    """Print debug message if debugging is switched on"""
    errormsg(what)
    return what


def do_nothing(*args, **kwargs):  # pylint: disable=unused-argument
    """A blank function to do nothing

    .. versionadded:: 1.1"""


def errormsg(msg):
    """Intended for end-user-visible error messages.

    (Currently just writes to stderr with an appended newline, but could do
    something better in future: e.g. could add markup to distinguish error
    messages from status messages or debugging output.)

    Note that this should always be combined with translation::

      import inkex
      ...
      inkex.errormsg(_("This extension requires two selected paths."))
    """
    try:
        sys.stderr.write(msg)
    except TypeError:
        sys.stderr.write(str(msg))
    except UnicodeEncodeError:
        # Python 2:
        # Fallback for cases where sys.stderr.encoding is not Unicode.
        # Python 3:
        # This will not work as write() does not accept byte strings, but AFAIK
        # we should never reach this point as the default error handler is
        # 'backslashreplace'.

        # This will be None by default if stderr is piped, so use ASCII as a
        # last resort.
        encoding = sys.stderr.encoding or "ascii"
        sys.stderr.write(msg.encode(encoding, "backslashreplace"))

    # Write '\n' separately to avoid dealing with different string types.
    sys.stderr.write("\n")


class AbortExtension(Exception):
    """Raised to print a message to the user without backtrace"""


class DependencyError(NotImplementedError):
    """Raised when we need an external python module that isn't available"""


class FragmentError(Exception):
    """Raised when trying to do rooty things on an xml fragment"""


def to(kind):  # pylint: disable=invalid-name
    """
    Decorator which will turn a generator into a list, tuple or other object type.
    """

    def _inner(call):
        def _outer(*args, **kw):
            return kind(call(*args, **kw))

        return _outer

    return _inner


def strargs(string, kind=float):
    """Returns a list of floats from a string

    .. versionchanged:: 1.1
        also splits at -(minus) signs by adding a space in front of the - sign

    .. versionchanged:: 1.2
        Full support for the `SVG Path data BNF
        <https://www.w3.org/Graphics/SVG/1.1/paths.html#PathDataBNF>`_
    """
    return [kind(val) for val in NUMBER_REX.findall(string)]


class classproperty:  # pylint: disable=invalid-name, too-few-public-methods
    """Combine classmethod and property decorators"""

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, owner):
        return self.func(owner)


def filename_arg(name):
    """Existing file to read or option used in script arguments"""
    filename = os.path.abspath(os.path.expanduser(name))
    if not os.path.isfile(filename):
        raise ArgumentTypeError(f"File not found: {name}")
    return filename


def pairwise(iterable, start=True):
    "Iterate over a list with overlapping pairs (see itertools recipes)"
    first, then = tee(iterable)
    starter = [(None, next(then, None))]
    if not start:
        starter = []
    return starter + list(zip(first, then))


def circular_pairwise(l):
    """Iterate over a list with overlapping pairs in a periodic way, i.e.
    [1, 2, 3] -> [(1, 2), (2, 3), (3, 1)]

    ..versionadded:: 1.3.1"""
    second = cycle(l)
    next(second)
    return zip(l, second)


EVAL_GLOBALS = {}
EVAL_GLOBALS.update(random.__dict__)
EVAL_GLOBALS.update(math.__dict__)


def math_eval(function, variable="x"):
    """Interpret a function string. All functions from math and random may be used.

    .. versionadded:: 1.1

    Returns:
        a lambda expression if sucessful; otherwise None.
    """
    try:
        if function != "":
            return eval(
                f"lambda {variable}: " + (function.strip('"') or "t"), EVAL_GLOBALS, {}
            )
    # handle incomplete/invalid function gracefully
    except SyntaxError:
        pass
    return None


def is_number(string):
    """Checks if a value is a number

    .. versionadded:: 1.2"""
    try:
        float(string)
        return True
    except ValueError:
        return False


def rational_limit(f: np.poly1d, g: np.poly1d, t0):
    """Computes the limit of the rational function (f/g)(t)
    as t approaches t0.

    .. versionadded:: 1.4"""
    assert g != np.poly1d([0])
    if g(t0) != 0:
        return f(t0) / g(t0)
    elif f(t0) == 0:
        return rational_limit(f.deriv(), g.deriv(), t0)
    else:
        raise ValueError("Limit does not exist.")


def callback_method(func):
    def notify(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        self._callback()
        return result

    return notify


class NotifyList(list):
    """A list that calls a callback after it is modified
    (to notify a parent about the modification).
    Modified from https://stackoverflow.com/a/13259435/3298143

    .. versionadded:: 1.4"""

    extend = callback_method(list.extend)
    append = callback_method(list.append)
    remove = callback_method(list.remove)
    pop = callback_method(list.pop)
    __delitem__ = callback_method(list.__delitem__)
    __setitem__ = callback_method(list.__setitem__)
    __iadd__ = callback_method(list.__iadd__)
    __imul__ = callback_method(list.__imul__)

    def __getitem__(self, item):
        """Ensure that slicing returns a list of the same datatype"""
        if isinstance(item, slice):
            return self.__class__(list.__getitem__(self, item))
        return list.__getitem__(self, item)

    def __init__(self, *args, callback=None):
        self.callback = None
        list.__init__(self, *args)
        self.callback = callback

    def _callback(self):
        if self.callback is not None:
            self.callback(self)

    def toggle(self, value):
        """If exists, remove it, if not, add it"""
        value = str(value)
        if value in self:
            return self.remove(value)
        return self.append(value)


class NotifyOrderedDict(OrderedDict):
    """An OrderedDict that notifies a callback after a value is changed

    .. versionadded:: 1.4"""

    clear = callback_method(OrderedDict.clear)
    popitem = callback_method(OrderedDict.popitem)
    update = callback_method(OrderedDict.update)
    setdefault = callback_method(OrderedDict.setdefault)
    __setitem__ = callback_method(OrderedDict.__setitem__)
    __delitem__ = callback_method(OrderedDict.__delitem__)

    def __init__(self, *args, callback=None, **kwargs):
        self.callback = None
        super().__init__(*args, **kwargs)
        self.callback = callback

    def _callback(self):
        if self.callback is not None:
            self.callback(self)

    def pop(self, key, default=None):
        super().pop(key, default)
        # On Python < 3.11, pop internally calls __delitem__.
        # This does not happen in 3.11. To avoid
        # calling the callback twice, we need to check the Python version.
        if sys.version_info >= (3, 11):
            if self.callback is not None:
                self.callback(self)


class NotifyDefaultDict(defaultdict):
    """A defaultdict that notifies a callback after a value is changed

    .. versionadded:: 1.4"""

    clear = callback_method(defaultdict.clear)
    popitem = callback_method(defaultdict.popitem)
    update = callback_method(defaultdict.update)
    setdefault = callback_method(defaultdict.setdefault)
    __setitem__ = callback_method(defaultdict.__setitem__)
    __delitem__ = callback_method(defaultdict.__delitem__)

    def __init__(self, *args, callback=None, **kwargs):
        self.callback = None
        super().__init__(*args, **kwargs)
        self.callback = callback

    def _callback(self):
        if self.callback is not None:
            self.callback(self)

    def pop(self, key, default=None):
        super().pop(key, default)
        # On Python < 3.11, pop internally calls __delitem__.
        # This does not happen in 3.11. To avoid
        # calling the callback twice, we need to check the Python version.
        if sys.version_info >= (3, 11):
            if self.callback is not None:
                self.callback(self)



"""
VNArtboard, Frame, VNLayer
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional

from .base import VNBaseElement
from .guide import VNGuideElement
from .styles import VNColor, VNGradient


@dataclass
class VNArtboard:
    """Represents Linearity Curve Artboard."""

    title: str
    frame: Frame
    layers: List[VNLayer]
    guides: Optional[List[VNGuideElement]]
    fillColor: Optional[VNColor]
    fillGradient: Optional[VNGradient]


@dataclass
class Frame:
    """Artboard frame."""

    width: float
    height: float
    x: float
    y: float


@dataclass
class VNLayer:
    """Represents Linearity Curve Layer."""

    name: str
    opacity: float
    isVisible: bool
    isLocked: bool
    isExpanded: bool
    elements: List[VNBaseElement]



#!/usr/bin/env python3
#
# Copyright 2008, 2009 Hannes Hochreiner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.
#
"""Effect to print out jessyInk summary"""

from collections import defaultdict
import inkex

from jessyink_install import JessyInkMixin, _


class Summary(JessyInkMixin, inkex.EffectExtension):
    """Print of jessyInk summary"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")

    def effect(self):
        self.is_installed()

        # Find the script node, if present
        for node in self.svg.xpath("//svg:script[@id='JessyInk']"):
            version = node.get("jessyink:version")
            if version:
                self.msg(_("JessyInk script version {} installed.".format(version)))
            else:
                self.msg(_("JessyInk script installed."))

        slides = []
        master_slide = None

        for node in self.svg.descendants().get(inkex.Layer):
            if node.get("jessyink:master_slide"):
                master_slide = node
            else:
                slides.append(node)

        if master_slide is not None:
            self.msg(_("\nMaster slide:"))
            self.describe_node(
                master_slide,
                "\t",
                ["<the number of the slide>", len(slides), "<the title of the slide>"],
            )

        for i, slide in enumerate(slides):
            self.msg(_("\nSlide {0!s}:").format(i + 1))
            self.describe_node(slide, "\t", [i + 1, len(slides), slide.label])

    def describe_node(self, node, prefix, dat):
        """Standard print out formatter"""

        self.msg(_(f"{prefix}Layer name: {node.label}"))
        self.describe_transition(node, prefix, "In")
        self.describe_transition(node, prefix, "Out")
        self.describe_autotext(node, prefix, dat)
        self.describe_effects(node, prefix)

    def describe_transition(self, node, prefix, transition):
        """Display information about transitions."""
        trans = inkex.Style(node.get(f"jessyink:transition{transition}"))
        if trans:
            name = trans["name"]
            if name != "appear" and "length" in trans:
                length = int(trans["length"] / 1000.0)
                self.msg(
                    _(
                        "{0}Transition {1}: {2} ({3!s} s)".format(
                            prefix, transition, name, length
                        )
                    )
                )
            else:
                self.msg(_(f"{prefix}Transition {transition}: {name}"))

    def describe_autotext(self, node, prefix, dat):
        """Display information about auto-texts."""
        auto_texts = {"slide_num": dat[0], "num": dat[1], "title": dat[2]}
        for x, child in enumerate(node.xpath(".//*[@jessyink:autoText]")):
            if not x:
                self.msg(_("\n{0}Auto-texts:").format(prefix))

            pid = child.getparent().get("id")
            val = auto_texts[child.get("jessyink:autoText")]
            self.msg(
                _(
                    _('{0}\t"{1}" (object id "{2}") will be replaced by "{3}".').format(
                        prefix, child.text, pid, val
                    )
                )
            )

    def describe_effects(self, node, prefix):
        """Display information about effects."""
        effects = sorted(self.collect_effects(node), key=lambda val: val[0])
        for x, (enum, effect) in enumerate(effects):
            ret = ""

            order = effect[0]["order"]
            if not x:
                ret += _("\n{0}Initial effect (order number {1}):").format(
                    prefix, order
                )
            else:
                ret += _("\n{0}Effect {1} (order number {2}):").format(
                    prefix, enum, order
                )

            for item in effect:
                eid = item["id"]
                if item["type"] == "view":
                    ret += _('{0}\tView will be set according to object "{1}"').format(
                        prefix, eid
                    )
                else:
                    ret += _('{0}\tObject "{1}"').format(prefix, eid)

                    if item["direction"] == "in":
                        ret += _(" will appear")
                    elif item["direction"] == "out":
                        ret += _(" will disappear")

                if item["name"] != "appear":
                    ret += _(' using effect "{0}"').format(item["name"])

                if "length" in item:
                    ret += _(" in {0!s} s").format(int(item["length"]) / 1000.0)

            self.msg(ret + ".\n")

    @staticmethod
    def collect_effects(node):
        """Collect information about effects."""
        effects = defaultdict(list)
        for child in node.xpath(".//*[@jessyink:effectIn]"):
            effect_data = inkex.Style(child.get("jessyink:effectIn"))
            effect_data["direction"] = "in"
            effect_data["id"] = child.get("id")
            effect_data["type"] = "effect"
            effects[effect_data["order"]].append(effect_data)

        for child in node.xpath(".//*[@jessyink:effectOut]"):
            effect_data = inkex.Style(child.get("jessyink:effectOut"))
            effect_data["direction"] = "out"
            effect_data["id"] = child.get("id")
            effect_data["type"] = "effect"
            effects[effect_data["order"]].append(effect_data)

        for child in node.xpath(".//*[@jessyink:view]"):
            effect_data = inkex.Style(child.get("jessyink:view"))
            effect_data["id"] = child.get("id")
            effect_data["type"] = "view"
            effects[effect_data["order"]].append(effect_data)
        return effects


if __name__ == "__main__":
    Summary().run()



#!/usr/bin/env python3
#
# Copyright (C) 2016 su_v, <suv-sf@users.sf.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Convert path to mesh gradient
"""

import inkex
from inkex.paths import Line, Curve
from inkex.elements import MeshGradient


class PathToMesh(inkex.EffectExtension):
    """Convert path data to mesh geometry."""

    def add_arguments(self, pars):
        pars.add_argument("--tab", help="The selected UI-tab")

    def add_mesh(self, meshgradient):
        """Add meshgradient definition to current document."""
        self.svg.defs.append(meshgradient)
        meshgradient.set_random_id("meshgradient")
        return meshgradient.get_id()

    def effect(self):
        """Main routine to convert path data to mesh geometry."""
        # loop through selection
        for node in self.svg.selection.filter(inkex.PathElement):
            csp = None
            meshgradient = None
            mesh_id = None
            # parse path data
            csp = node.path.to_superpath()
            # convert csp to meshgradient definition
            if csp is not None:
                # TODO: check length of path data / csp
                meshgradient = self.to_mesh(node, csp)
            # add meshgradient to document
            if meshgradient is not None:
                mesh_id = self.add_mesh(meshgradient)
            # apply meshgradient to node
            if mesh_id is not None:
                node.style["fill"] = "url(#{})".format(mesh_id)

    def to_mesh(self, node, csp, subpath=0):
        """Convert csp to meshgradient geometry."""
        # mesh data
        corners, edges = self.to_meshdata(csp[subpath])
        # alternating stop colors
        colors = [node.style.get("fill"), "#ffffff"]
        # define meshgradient with first corner as initial point
        meshgradient = MeshGradient.new_mesh(pos=corners[0], rows=1, cols=1)
        # define stops (stop-color, path) for first meshpatch
        meshgradient[0][0].stops(edges[0:4], colors)
        return meshgradient

    @staticmethod
    def to_meshdata(subpath):
        """Convert csp subpath to corners, edge path data."""
        if len(subpath) >= 5:
            corners = []
            edges = []
            for i, corner in enumerate(subpath[:4]):
                corners.append(corner[1])
                edge = [list(subpath[i]), list(subpath[i + 1])]
                edge[0][0] = list(edge[0][1])
                edge[1][2] = list(edge[1][1])
                if inkex.CubicSuperPath.is_line(edge[0], edge[1]):
                    edges.append(Line(*edge[1][1]))
                else:
                    edges.append(Curve(*(edge[0][2] + edge[1][0] + edge[1][1])))

        return corners, edges


if __name__ == "__main__":
    PathToMesh().run()



# This file is derived from
# https://github.com/avibrazil/NSKeyedUnArchiver
# License: LGPL3

import copy
import datetime
import plistlib
import struct
from typing import Dict, List


def _unserialize(
    o: dict, serialized: dict, removeClassName: bool, plist_top: bool = True
):
    if plist_top:
        reassembled = copy.deepcopy(o)
    else:
        reassembled = o

    finished = False
    while not finished:
        finished = True

        cursor = None
        if isinstance(reassembled, bytes):
            try:
                return NSKeyedUnarchiver(reassembled)
            except:
                # Not plist data, just plain binary
                return reassembled
        elif isinstance(reassembled, dict):
            cursor = reassembled.keys()
        elif isinstance(reassembled, list):
            cursor = range(len(reassembled))
        else:  # str, int etc
            print("reassembled is a " + str(type(reassembled)) + ":" + str(reassembled))
            return reassembled

        for k in cursor:
            #  print(f"cursor={k}")
            # UIDs references items in "serialized" ($objects)
            if isinstance(reassembled[k], plistlib.UID):
                reassembled[k] = copy.deepcopy(serialized[reassembled[k].data])

                # $null gets replaced by None
                if str(reassembled[k]) == "$null":
                    reassembled[k] = None

                finished = False

            elif isinstance(reassembled[k], dict) or isinstance(reassembled[k], list):
                reassembled[k] = _unserialize(
                    reassembled[k], serialized, removeClassName, plist_top=False
                )

                if (
                    "$class" in reassembled[k]
                    and "$classes" in reassembled[k]["$class"]
                ):
                    # Specialized handler for common class types
                    if "NSArray" in reassembled[k]["$class"]["$classes"]:
                        reassembled[k] = reassembled[k]["NS.objects"]

                    elif any(
                        c in reassembled[k]["$class"]["$classes"]
                        for c in ["NSMutableDictionary", "NSDictionary"]
                    ):
                        reassembled[k] = dict(
                            zip(reassembled[k]["NS.keys"], reassembled[k]["NS.objects"])
                        )

                    elif any(
                        c in reassembled[k]["$class"]["$classes"]
                        for c in ["NSMutableString", "NSString"]
                    ):
                        reassembled[k] = reassembled[k]["NS.string"]

                    elif any(
                        c in reassembled[k]["$class"]["$classes"]
                        for c in ["NSMutableData", "NSData"]
                    ):
                        reassembled[k] = reassembled[k]["NS.data"]

                    elif "NSDate" in reassembled[k]["$class"]["$classes"]:
                        apple2001reference = datetime.datetime(
                            2001, 1, 1, tzinfo=datetime.timezone.utc
                        )
                        reassembled[k] = datetime.datetime.fromtimestamp(
                            reassembled[k]["NS.time"] + apple2001reference.timestamp(),
                            datetime.timezone.utc,
                        )
                    if removeClassName and isinstance(reassembled[k], dict):
                        # Remove visual polution
                        if "$class" in reassembled[k]:  # check if it's there
                            del reassembled[k]["$class"]

                finished = True
            else:
                # strings will remain unchanged.
                pass

    return reassembled


def _decode_attrib_info(attr_bytes: bytes) -> List[Dict]:
    """
    Tries to decode NSAttributeInfo in KeyedArchived NSAttributedString.
    """
    # FIXME decode_attrib_info(legacy text) does not work for longer texts
    attr_data = []
    offset = 0
    while offset < len(attr_bytes):
        # 2bytes()
        if offset + 2 <= len(attr_bytes):
            length, attr_id = struct.unpack_from("<BB", attr_bytes, offset)
            attr_data.append(
                {
                    "length": length,
                    "attribute_id": attr_id,
                }
            )
            offset += 2
        else:
            break

    return attr_data


def NSKeyedUnarchiver(plist, removeClassName=True):
    """
    plist can be:
    • bytes      ⟹ pass it through plistlib.loads()
    • dict       ⟹ unserialize
    """
    # removed other formats as they are not needed

    if isinstance(plist, bytes):
        plistdata = plistlib.loads(plist)
    elif isinstance(plist, dict):
        # plist is already a plistlib-parsed dict
        plistdata = plist
    else:
        raise TypeError(
            "Trying to plist-parse something that is neither a PurePath, file name, XML text, plist bytes stream nor a dict."
        )

    if "$top" in plistdata:
        o = copy.deepcopy(plistdata["$top"])
        unserialized = _unserialize(o, plistdata["$objects"], removeClassName)
    else:
        raise TypeError("Passed object is not an NSKeyedArchiver")

    if len(unserialized) == 1 and "root" in unserialized:
        # Unserialized data contains only 1 object, so no need to nest it under 'root'
        unserialized = unserialized["root"]

    if unserialized.get("NSAttributeInfo") is not None:
        attr_info = unserialized["NSAttributeInfo"]
        attr_bytes = attr_info if isinstance(attr_info, bytes) else bytes(attr_info)
        unserialized["NSAttributeInfo"] = _decode_attrib_info(attr_bytes)

    return unserialized



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Utilities for conversion"""

import inkex

from ..xamlobjects import XAMLObject


def convert_attributes(source, target: XAMLObject, definitions):  # convert with dict
    """Extracts attribute from source and writes it to target by using definition.
    Definition types:
    * lambda(str) -> str
    * Dict map
    * List -> capitalizes entry
    4th entry of definition: default XAML value
    5th entry of definition: default SVG value
    """
    for dfn in definitions:
        val = None
        if isinstance(source, inkex.Style):
            val = source(dfn[0])
        elif source is not None:
            val = source.get(dfn[0], None)
        if val is None and len(dfn) == 5:
            # set to the default SVG value
            val = dfn[4]
        if val is None:
            continue
        if dfn[0] == "stroke-dasharray" and val == []:
            continue
        result: str = ""
        if isinstance(dfn[2], dict):
            # This should be always true, by the EnumValue construct in properties
            # handling
            if val in dfn[2]:
                result = dfn[2][val]
            elif len(dfn) == 5:
                result = dfn[2][dfn[4]]
        if isinstance(dfn[2], list):
            if val in dfn[2] and isinstance(val, str):
                result = val.capitalize()
        if callable(dfn[2]):
            result = dfn[2](val)
        target.add_property(dfn[1], result, None if len(dfn) < 4 else dfn[3])



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from dataclasses import dataclass
import math
from typing import List, Optional

import inkex

from ..parser.types import AFObject, Curve18


@dataclass
class AFCurve:
    curves: List[List[Curve18]]
    closed: List[bool]

    @classmethod
    def from_af(cls, crvs: AFObject) -> AFCurve:
        assert crvs.get_type() in ["PCvD", "CrvD"], f"Not a curve {crvs}"
        data = crvs["Data"]

        _byte: int = data[0]
        size: Optional[int] = data[1] if crvs.get_type() == "PCvD" else None
        curv_ptr = 1 if size is None else 2
        if size is None:
            size = (len(data) - curv_ptr) // 2

        if _byte != 0:
            raise NotImplementedError(f"_byte {_byte} is not implemented\n{crvs}")

        curves: List[List[Curve18]] = []
        closeds: List[bool] = []
        for curv_idx in range(size):
            closed = data[curv_idx * 2 + curv_ptr]
            curve = data[curv_idx * 2 + curv_ptr + 1]
            assert isinstance(closed, bool), f"Invalid curve\n{data}"
            assert isinstance(curve, list), f"Invalid curve\n{data}"
            assert not curve or isinstance(curve[0], Curve18), (
                f'Invalid curve type "{type(curve[0])}" in \n{data}'
            )

            curves.append(curve)
            closeds.append(closed)

        return cls(curves, closeds)

    @staticmethod
    def _sort_nodes(nodes: List[Curve18], n: int) -> None:
        for i in range(0, len(nodes), n):
            nodes[i : i + n] = sorted(nodes[i : i + n], key=lambda p: p.index)

    def get_svg_path(self) -> inkex.Path:
        path = inkex.Path()
        for curve, closed in zip(self.curves, self.closed):
            # Paths sometimes start with a (symmetric) invisible handle on the first node,
            # we skip that. In that case the outgoing handle also has
            # index: 2 = incoming handle, 0 = node, 1 = outgoing handle
            i = 0
            while curve[i].index != 0:
                i += 1

            path.append(inkex.paths.Move(*curve[i].point))
            i += 1
            while i < len(curve):
                p1 = curve[i]
                if p1.index == 0:
                    path.append(inkex.paths.Line(*p1.point))
                    i += 1
                else:
                    if i + 2 >= len(curve):
                        break  # outgoing (symmetric) handle of the last node
                    last = path[-1].cend_point(0j, 0j)
                    p2 = curve[i + 1]
                    p3 = curve[i + 2]
                    assert p1.index == 1 and p2.index == 2 and p3.index == 0
                    if (
                        math.isclose(last.real, p1.point[0])
                        and math.isclose(last.imag, p1.point[1])
                        and math.isclose(p2.point[0], p3.point[0])
                        and math.isclose(p2.point[1], p3.point[1])
                    ):
                        path.append(inkex.paths.Line(*p3.point))
                    else:
                        path.append(inkex.paths.Curve(*p1.point, *p2.point, *p3.point))

                    i += 3

            if closed:
                path.append(inkex.paths.ZoneClose())

        return path



#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
# Copyright 2022 Simon Duerr <dev@simonduerr.eu>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import hashlib
from import_sources import RemoteSource


class Wikimedia(RemoteSource):
    name = "Wikimedia"
    icon = "sources/wikimedia.svg"
    base_url = "https://commons.wikimedia.org/w/api.php"

    def search(self, query):
        params = {
            "action": "query",
            "format": "json",
            "uselang": "en",
            "generator": "search",
            "gsrsearch": "filetype:bitmap|drawing filemime:svg " + query,
            "gsrlimit": 40,
            "gsroffset": 0,
            "gsrinfo": "totalhits|suggestion",
            "gsrprop": "size|wordcount",
            "gsrnamespace": 6,
            "prop": "info|imageinfo|entityterms",
            "inprop": "url",
            "iiprop": "url|size|mime|user|extmetadata",
            "iiurlheight": 180,
            "wbetterms": "label",
        }
        pages = []
        try:
            response = self.session.get(self.base_url, params=params).json()
            if "error" in response:
                raise IOError(response["error"]["info"])
            pages = response["query"]["pages"].values()
        except:
            pass

        for item in pages:
            img = item["imageinfo"][0]
            # get standard licenses
            # for non standard licenses we have to get the ShortName and provide the url to the resource
            try:
                license = img["extmetadata"]["License"]["value"]
                if license in ["cc0", "pd"]:
                    license = "cc-0"
            except KeyError:
                license = img["extmetadata"]["LicenseShortName"]["value"]
            yield {
                "id": item.get("pageid", None),
                "name": item["title"].split(":", 1)[-1],
                "author": img["user"],
                "license": license,
                "summary": "",  # No data
                "thumbnail": img["thumburl"],
                "created": item["touched"],
                "descriptionurl": item["canonicalurl"],
                "popularity": 0,  # No data
                "file": img["url"],
            }



"""
Definition of available Inkscape target directories and their machinery
"""

from .target import ExtensionsTarget, BasicTarget

TARGETS = [
    # Website slug, Visible name, local directory, search instead of list
    ExtensionsTarget(
        "extension", "Extensions", "extensions", True, filters=("*.inx",)
    ),
    BasicTarget(
        "template", "Templates", "templates", True, filters=("*.svg",)
    ),
    BasicTarget("palette", "Shared Paletts", "palettes", filters=("*.gpl",)),
    BasicTarget("symbol", "Symbol Collections", "symbols", filters=("*.svg",)),
    BasicTarget("keyboard", "Keyboard Shortcuts", "keys", filters=("*.xml",)),
    # ('marker', 'Marker Collections', '', False), # No marker config
    # ('pattern', 'Pattern Collections', '', False), # No pattern config
    # ('', 'User Interface Themes', 'themes', False), # No website category
    # ('', 'Paint Server', 'paint', False), # No website category
    # ('', 'User Interfaces', 'ui', False), # No website category
    # ('', 'Icon Sets', 'icons', False), # No website category
]



# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""A Python path turtle for Inkscape extensions"""

import math
import random
from typing import List, Union

from .paths import Line, Move, Path, PathCommand
from .elements import PathElement, Group, BaseElement
from .styles import Style


class PathTurtle:
    """A Python path turtle

    .. versionchanged:: 1.2
        pTurtle has been renamed to PathTurtle."""

    def __init__(self, home=(0, 0)):
        self.__home = [home[0], home[1]]
        self.__pos = self.__home[:]
        self.__heading = -90
        self.__path = ""
        self.__draw = True
        self.__new = True

    def forward(self, mag: float):
        """Move turtle forward by mag in the current direction."""
        self.setpos(
            (
                self.__pos[0] + math.cos(math.radians(self.__heading)) * mag,
                self.__pos[1] + math.sin(math.radians(self.__heading)) * mag,
            )
        )

    def backward(self, mag):
        """Move turtle backward by mag in the current direction."""
        self.setpos(
            (
                self.__pos[0] - math.cos(math.radians(self.__heading)) * mag,
                self.__pos[1] - math.sin(math.radians(self.__heading)) * mag,
            )
        )

    def right(self, deg):
        """Rotate turtle right by deg degrees.

        Changed in inkex 1.2: The turtle now rotates right (previously left) when
            calling this method."""
        self.__heading += deg

    def left(self, deg):
        """Rotate turtle left by deg degrees.

        Changed in inkex 1.2: The turtle now rotates left (previously right) when
            calling this method."""
        self.__heading -= deg

    def penup(self):
        """Enable non-drawing / moving mode"""
        self.__draw = False
        self.__new = False

    def pendown(self):
        """Enable drawing mode"""
        if not self.__draw:
            self.__new = True
        self.__draw = True

    def pentoggle(self):
        """Switch between drawing and moving mode"""
        if self.__draw:
            self.penup()
        else:
            self.pendown()

    def home(self):
        """Move to home position"""
        self.setpos(self.__home)

    def clean(self):
        """Delete current path"""
        self.__path = ""

    def clear(self):
        """Delete current path and move to home"""
        self.clean()
        self.home()

    def setpos(self, arg):
        """Move/draw to position, depending on the current state"""
        if self.__new:
            self.__path += "M" + ",".join([str(i) for i in self.__pos])
            self.__new = False
        self.__pos = arg
        if self.__draw:
            self.__path += "L" + ",".join([str(i) for i in self.__pos])

    def getpos(self):
        """Returns the current position"""
        return self.__pos[:]

    def setheading(self, deg):
        """Set the heading to deg degrees"""
        self.__heading = deg

    def getheading(self):
        """Returns the heading in degrees"""
        return self.__heading

    def sethome(self, arg):
        """Set home position"""
        self.__home = list(arg)

    def getPath(self):
        """Returns the current path"""
        return self.__path

    def rtree(self, size, minimum, pt=False):
        """Generates a random tree"""
        if size < minimum:
            return
        self.fd(size)
        turn = random.uniform(20, 40)
        self.rt(turn)
        self.rtree(size * random.uniform(0.5, 0.9), minimum, pt)
        self.lt(turn)
        turn = random.uniform(20, 40)
        self.lt(turn)
        self.rtree(size * random.uniform(0.5, 0.9), minimum, pt)
        self.rt(turn)
        if pt:
            self.pu()
        self.bk(size)
        if pt:
            self.pd()

    # pylint: disable=invalid-name
    fd = forward
    bk = backward
    rt = right
    lt = left
    pu = penup
    pd = pendown


pTurtle = PathTurtle  # should be deprecated


class PathBuilder:
    """This helper class can be used to construct a path and insert it into a
    document.

    .. versionadded:: 1.2"""

    def __init__(self, style: Style):
        """Initializes a PathDrawHelper object

        Args:
            style (Style): Style of the path.
        """
        self.current = Path()
        self.style = style

    def add(self, command: Union[PathCommand, List[PathCommand]]):
        """Add a Path command to the Helper

        Args:
            command (Union[PathCommand, List[PathCommand]]): A (list of) PathCommand(s)
                to be appended.
        """
        if isinstance(command, list):
            self.current.extend(command)
        else:
            self.current.append(command)

    def terminate(self):
        """Terminates current subpath. This method does nothing by default and is
        supposed to be overridden in subclasses."""

    def append_next(self, sibling_before: BaseElement):
        """Insert the resulting Path as :class:`inkex.elements._polygons.PathElement`
        into the document tree.

        Args:
            sibling_before (BaseElement): The element the resulting path will be
                appended after.
        """
        pth = PathElement()
        pth.path = self.current
        pth.style = self.style
        sibling_before.addnext(pth)

    def Move_to(self, x, y):  # pylint: disable=invalid-name
        """Shorthand to insert an absolute move command: `M x y`.

        Args:
            x (Float): x coordinate to move to
            y (Float): y coordinate to move to
        """
        self.add(Move(x, y))

    def Line_to(self, x, y):  # pylint: disable=invalid-name
        """Shorthand to insert an absolute lineto command: `L x y`.

        Args:
            x (Float): x coordinate to draw a line to
            y (Float): y coordinate to draw a line to
        """
        self.add(Line(x, y))


class PathGroupBuilder(PathBuilder):
    """This helper class can be used to construct a group of paths that all have the
    same style.

    .. versionadded:: 1.2"""

    def __init__(self, style):
        super().__init__(style)
        self.result = Group()

    def terminate(self):
        """Terminates the current Path, and appends it to the group if it is not
        empty."""
        if len(self.current) > 1:
            pth = PathElement()
            pth.path = self.current.to_absolute()
            pth.style = self.style
            self.result.append(pth)
        self.current = Path()

    def append_next(self, sibling_before: BaseElement):
        """Insert the resulting Path as :class:`inkex.elements._groups.Group` into the
            document tree.

        Args:
            sibling_before (BaseElement): The element the resulting group will be
                appended after.
        """
        sibling_before.addnext(self.result)



# SPDX-FileCopyrightText: 2023 Software Freedom Conservancy <info@sfconservancy.org>
# SPDX-FileCopyrightText: 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Script to extract afdesign data.

poetry run python -m inkaf.extract file.afdesign ...

Stores full extracted data in *.afdesign.json

"""

from io import BytesIO
import os
import json
import argparse
from pathlib import Path
import shutil
import traceback
import sys
from typing import Any, Dict
import logging

from inkaf.svg.convert import AFConverter

HERE = os.path.dirname(__file__) or "."
# This is suggested by https://docs.python-guide.org/writing/structure/.
sys.path.insert(0, os.path.abspath(os.path.join(HERE, "..")))

from inkaf.parser.parse import AFParser  # noqa: E402
from inkaf.parser.json_encoder import EnhancedJSONEncoder  # noqa: E402
from inkaf.parser.types import AFObject  # noqa: E402
from inkaf.parser.extract import AFExtractor, FATFile, FATFileType, NormalFATFile  # noqa: E402
from inkaf.parser.consts import content_magic  # noqa: E402

EXTRACTION_ENDING = ".json"

parser = argparse.ArgumentParser(
    prog="poetry run python3 -m inkaf.extract",
    description="Inkscape AI extraction tool: Unzips and additionally extracts (preprocessed) parts of the passed AI files.",
)

parser.add_argument("filename", nargs="*", help="One or more afdesign files")
parser.add_argument(
    "-n", "--if-newer", action="store_true", help="Only re-extract files if newer"
)
parser.add_argument(
    "-v", "--verbose", action="store_true", help="verbose output on errors"
)
parser.add_argument(
    "-z", "--extract-full", action="store_true", help="extract entire archive"
)
parser.add_argument(
    "-c",
    "--collect-unit-tests",
    action="store_true",
    help="Collect unit tests. Modify the wrapper function to decide what to collect",
)
parser.add_argument(
    "-e",
    "--no-enhance",
    action="store_true",
    help="Do not enhance the extracted svg",
)


args = parser.parse_args()

typetests: Dict[Any, Any] = {}


def create_type_test():
    # We monkey-patch the parser to extract unit test data.

    _load_field = AFParser.load_field

    def new_load_field(self: AFParser, type_, array):
        start = self.get_pos()
        result = _load_field(self, type_, array)
        end = self.get_pos()

        type_ = type_ + (128 if array else 0)
        do = False

        if isinstance(result, AFObject) and result.get_type() in (
            "BGrd",
            "GRAY",
        ):
            do = True

        if do:
            self.data.stream.seek(start - 5)
            data = self.data.stream.read(end - start + 5)
            typetests[result.get_type()] = (data, result)

        return result

    AFParser.load_field = new_load_field  # type: ignore


if args.collect_unit_tests:
    create_type_test()


def process_fat_file(extractor: AFExtractor, fat_dir: Path, file: FATFile):
    if isinstance(file, NormalFATFile) and file.flag == FATFileType.NORMAL_WITH_NAME:
        try:
            data = extractor.extract(file)
        except Exception as e:
            print(
                f"\tbroken extraction: {traceback.format_exc() if args.verbose else e}"
            )
            return True
        # Store the .dat
        assert file.name is not None
        target = fat_dir / (file.name + ".dat")
        target.parent.mkdir(exist_ok=True)
        with open(target, "wb") as wf:
            wf.write(data)
        if data.startswith((content_magic).to_bytes(4, byteorder="little")):
            # Try to extract the json
            try:
                parsed = AFParser(BytesIO(data)).parse()
                with open(fat_dir / (file.name + ".json"), "w") as f:
                    json.dump(parsed, f, cls=EnhancedJSONEncoder, indent=4)
                print(
                    f"\t{file.name}: Parsed content written to {fat_dir / (file.name + '.json')}"
                )
            except Exception as e:
                print(
                    f"\tbroken binary parsing: {traceback.format_exc() if args.verbose else e}"
                )
                return True
    return False


def extract(file_name: Path, write_archive=args.extract_full) -> AFExtractor:
    stream = open(file_name, "rb")
    error = False
    extractor = AFExtractor(stream)
    assert extractor.content is not None

    print("\tFile version", extractor.content.header.version)
    print("\tFile Access Tables: ", len(extractor.fats), extractor.fats)

    if write_archive:
        archive_folder = file_name.parent / file_name.stem
        if archive_folder.exists():
            shutil.rmtree(archive_folder)
        for fat in extractor.fats:
            fat_dir = archive_folder / str(fat.creation_date).replace(":", "_")
            fat_dir.mkdir(parents=True, exist_ok=True)
            for file in fat.files:
                e = process_fat_file(extractor, fat_dir, file)
                error = e or error
    if error:
        raise ValueError("One or more errors occurred during extraction")
    return extractor


def process_file(file_name):
    if file_name.suffix != ".afdesign":
        return False, False

    print("Found file:", file_name)
    output_file_name = file_name.with_suffix(".afdesign" + EXTRACTION_ENDING)
    if output_file_name.exists() and args.if_newer:
        if file_name.stat().st_mtime > output_file_name.stat().st_mtime:
            print(f"\t{output_file_name} is old, replace")
        else:
            print(
                f"\tskip: {output_file_name} exists and is newer than original document."
            )
            return False, False
    try:
        extractor = extract(file_name)
        result = process_fat_file(
            extractor, file_name.parent, extractor.get_head_revision("doc.dat")
        )
        if result:
            return True, True

        print(f"\tnew: {output_file_name}")

    except Exception as e:
        print(f"\tbroken extraction: {traceback.format_exc() if args.verbose else e}")
        print("\tfail:", file_name)
        return True, True

    try:
        with open(file_name, "rb") as stream:
            converter = AFConverter()
            converter.convert(AFExtractor(stream), not args.no_enhance)
            res = converter.doc.getroot().tostring()
            newfile = file_name.with_suffix(".svg")
            if res != "":
                with open(newfile, "w") as f:
                    f.write(res.decode())

            print(f"\tWrote parsed document to {newfile}")
    except Exception as e:
        print(f"\tbroken conversion: {traceback.format_exc() if args.verbose else e}")
        print("\tfail:", file_name)
        return True, True

    return False, True


def main(files=args.filename):
    """Extract the files."""

    logging.basicConfig(level=logging.INFO if args.verbose else logging.ERROR)

    did_something = False
    error = False
    broken = []

    def _process(path):
        e_, d_ = process_file(path)

        if e_:
            broken.append(path)
        return did_something or d_, error or e_, broken

    for file_name in map(Path, files):
        if file_name.is_dir():
            print(os.listdir(file_name))
            for fn in os.listdir(file_name):
                did_something, error, broken = _process(file_name / fn)
        else:
            did_something, error, broken = _process(file_name)

    if not did_something:
        print("Nothing to do!")
    if error:
        print(
            "We could not process at least one file. Isolate the problem and create a unit test."
        )
        print("Broken files:\n\t", "\n\t".join(map(str, broken)))
    if args.collect_unit_tests:
        print(
            "\n".join(
                str(typetests[key]) + f", #{key}" for key in sorted(typetests.keys())
            )
        )

        print("Located:", sorted(set(typetests.keys())))
    return int(error)


if __name__ == "__main__":
    exit(main())



# coding=utf-8
#
# Copyright(C) 2021 -  Windell H. Oskay, www.evilmadscientist.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""
Hershey Text 3.0.8, 2023-10-22

Copyright 2023, Windell H. Oskay, www.evilmadscientist.com

Major revisions in Hershey Text 3.0:

1. Migrate font format to use SVG fonts.
    - SVG fonts support unicode, meaning that we can use a full range of
        characters. We are no longer limited to the ASCII range that the
        historical Hershey font formats used.
    - Arbitrary curves are supported within glyphs; we are no longer limited to
        the straight line segments used in the historical Hershey format.
    - The set of fonts can now be expanded.

2. Add a mechanism for adding your own SVG fonts, either within the
    folder containing the default fonts, or from an external file or directory.
    This is particularly important for installations where one does not
    have access to edit the contents of the Inkscape extensions directory.

3. Support font mapping: If a given font face is used for a given block of
    text, check first to see if a matching SVG font is present. If not,
    substitute with the default (selected) stroke font from the list of
    included fonts.

4. Instead of entering text (one line at a time) in the extension,
    this script now converts text (either all text, or all selected text)
    in the document, replacing it in place. While not every possible
    method of formatting text is supported, many are.

"""

import os
import math

from copy import deepcopy

import inkex
from inkex import Transform, Style, units, AbortExtension
from inkex.localization import inkex_gettext as _

from inkex import (
    load_svg,
    Group,
    TextElement,
    FlowPara,
    SVGfont,
    FontFace,
    FlowSpan,
    Glyph,
    MissingGlyph,
    Tspan,
    FlowRoot,
    Rectangle,
    Use,
    PathElement,
    Defs,
)


class Hershey(inkex.Effect):
    """
    An extension for use with Inkscape 1.0
    """

    def __init__(self):
        super(Hershey, self).__init__()

        self.arg_parser.add_argument(
            "--tab",
            dest="mode",
            default="render",
            help="The active tab or mode when Apply was pressed",
        )

        self.arg_parser.add_argument(
            "--fontface",
            dest="fontface",
            default="HersheySans1",
            help="The selected font face when Apply was pressed",
        )

        self.arg_parser.add_argument(
            "--otherfont",
            dest="otherfont",
            default="",
            help="Optional other font name or path to use",
        )

        self.arg_parser.add_argument(
            "--preserve",
            type=inkex.Boolean,
            dest="preserve_text",
            default=False,
            help="Preserve original text",
        )

        self.arg_parser.add_argument(
            "--action",
            dest="util_mode",
            default="sample",
            help="The utility option selected",
        )

        self.arg_parser.add_argument(
            "--text",
            dest="sample_text",
            default="\nThe Quick Brown Fox Jumps Over a Lazy Dog",
            help="Text to use for font table",
        )

        self.font_file_list = dict()
        self.font_load_fail = False

        self.svg_height = None
        self.svg_width = None

        self.output_generated = False

        self.warn_unflow = False
        self.warn_unkern = False
        self.warn_textpath = (
            False  # For future use: Give warning about text attached to path.
        )
        self.font_dict = dict()  # Font dictionary - Dictionary of loaded fonts

        self.nodes_to_delete = []  # List of font elements to remove

        self.vb_scale_factor = 0.0104166666

        self.text_string = ""
        self.text_families = []  # List of font family for characters in the string
        self.text_heights = []  # List of font heights
        self.text_spacings = []  # List of vertical line heights
        self.text_aligns = []  # List of horizontal alignment values
        self.text_x = []  # List; x-coordinate of text line start
        self.text_y = []  # List; y-coordinate of text line start
        self.line_number = 0
        self.new_line = True
        self.render_width = 1

    PX_PER_INCH = 96.0

    help_text = """====== Hershey Text Help ======

The Hershey Text extension is designed to replace text in your document (either
selected text or all text) with specialized "stroke" or "engraving" fonts
designed for plotters.

Whereas regular "outline" fonts (e.g., TrueType) work by filling in the region
inside an invisible outline, stroke fonts are composed only of individual lines
or strokes with finite width; much like human handwriting when using a physical
pen.

Stroke fonts are most often used for creating text-like paths that computer
controlled drawing and cutting machines (from pen plotters to CNC routers) can
efficiently follow.

A full user guide for Hershey Text is available to download from
    http://wiki.evilmadscientist.com/hershey


   ==== Basic operation ====

To use Hershey Text, start with a document that contains text objects. Select
the "Render" tab of Hershey Text, and choose a font face from the pop-up menu.

When you click Apply, it will render all text elements on your page with the
selected stroke-based typeface. If you would like to convert only certain text
elements, click Apply with just those elements selected.

If the "Preserve original text" box is checked, then the original text elements
on the page will be preserved even when you click Apply. If it is unchecked,
then the original font elements will be removed once rendered.

You can generate a list of available SVG fonts or a list of all glyphs available
in a given font by using the tools available on the "Utilities" tab.


   ==== How Hershey Text works ====

Hershey Text works by performing font substitution, starting with the text in
your document and replacing it with paths generated from the characters in the
selected SVG font.

Hershey Text uses fonts in the SVG font format. While SVG fonts are one of the
few types that support stroke-based characters, it is important to note that
converting an outline font to SVG format does not convert it to a stroke based
font. Indeed, most SVG fonts are actually outline fonts.

This extension *does not* convert outline fonts into stroke fonts, nor does it
convert other fonts into SVG format. Its sole function is to replace the text
in your document with paths from the selected SVG font.


   ==== Using an external SVG font ====

To use an external SVG font -- one not included with the distribution -- select
"Other" for the name of the font in the pop-up menu on the "Render" tab. Then,
do one of the following:

(1) Add your SVG font file (perhaps "example.svg") to the "svg_fonts" directory
within your Inkscape extensions directory, and enter the name of the font
("example") in the "Other SVG font name or path" box on the "Render" tab.

or

(2) Place your SVG font file anywhere on your computer, and enter the full path
to the file  in the "Other SVG font name or path" box on the "Render" tab.
A full path might, for example, look like:
    /Users/Robin/Documents/AxiDraw/fonts/path_handwriting.svg


   ==== Using SVG fonts: Advanced methods ====

In addition to using a single SVG font for substitution, you can also use
font name mapping to automatically use particular stroke fonts in place of
specific font faces, to support various automated workflows and to support
the rapid use of multiple stroke font faces within the same document.

Several SVG fonts are included with this distribution, including both
single-stroke and multi-stroke fonts. These fonts are included within the
"svg_fonts" directory within your Inkscape extensions directory.

You can select the font that you would like to use from the pop-up menu on the
"Render" Tab. You can also make use of your own SVG fonts.

Order of preference for SVG fonts:

(1) If there is an SVG font with name matching that of the font for a given
piece of text, that font will be used. For example, if the original text is in
font "FancyScript" and there is a file in svg_fonts with name FancyScript.svg,
then FancyScript.svg will be used to render the text.

(2) Otherwise (if there is no SVG font available matching the name of the font
for a given block of text), the face selected from the "Font face" pop-up menu
will be used as the default font when rendering text with Hershey Text.

(3) You can also enter text in the "Name/Path" box, which can represent one of
the following: (i) a font name (for a font located in the svg_fonts directory),
(ii) the path to a font file elsewhere on your computer, or (iii) the path to a
directory containing (one or more) font files.

(3a) Using a font name:
If you move a custom SVG font file into your svg_fonts directory, then you can
enter the name of the SVG font in the "Name/Path" text box and select "Other"
from the pop-up menu. Then, the named font will be used as the default.

(3b) Using a file path:
If you enter the path to an SVG font file in the "Name/Path" text box and
select "Other" from the pop-up menu. Then, that font will be used as the
default. All SVG fonts located in the same directory as that font file will
also be available for name-based font substitution. If there are multiple
font-name matches, files in an external directory take precedence over ones in
the svg_fonts directory.

(3c) Using a directory path:
If you enter the path to a directory containing SVG font files in the
"Name/Path" text box, then all SVG font files files in that directory will be
available for name-based font substitution. If there are multiple font-name
matches, files in an external directory take precedence over ones in the
svg_fonts directory.



Tips about using these methods with your own custom fonts:

(A) These methods can be used to render different text elements with different
SVG font faces. You can even rename a font -- either your own custom one or one
of the bundled ones -- to match the name of a font that you're using. For
example, if you rename a script font you name a font to "Helvetica.svg",
then all text in Helvetica will be replaced with that SVG font.

(B) Using a directory path (3c) is a particularly helpful method if you do
not have access to modify the svg_fonts directory.



   ==== Limitations ====

This extension renders text into non-editable paths, generated from the
character geometry of SVG fonts. Once you have rendered the text, the resulting
paths can be edited with path editing tools, but not text editing tools.

Since this extension works by a process of font substitution, text spanning a
single line will generally stay that way, whereas text flowed in a box (that
may span multiple lines) will be re-flowed from scratch. Style information such
as text size and line spacing can be lost in some cases.

We recommend that you use the live preview option to achieve best results with
this extension.


(c) 2021 Windell H. Oskay
Evil Mad Scientist Laboratories
"""

    def getlength_inch(self, name):
        """
        Get the <svg> attribute with name "name", and parse it as a length,
        into a value and associated units. Return value in inches.

        This may cause scaling issues in some circumstances. Note, for
        example, that Adobe Illustrator uses 72 px per inch, and Inkscape
        used 90 px per inch prior to version 0.92.
        """
        string_to_parse = self.document.getroot().get(name)
        if string_to_parse:
            value, unit = units.parse_unit(string_to_parse)
            if value is None:
                return None
            bad_units = {"%", "ex", "em"}  # Unsupported units
            if unit in bad_units:
                return None

            return units.convert_unit(string_to_parse, "in")
        return None

    def units_to_userunits(self, input_string):
        """
        Custom replacement for the old "unittouu" routine

        Parse the attribute into a value and associated units.
        Return value in user units (typically "px").
        Importantly, return None for malformed inputs.
        """

        value, _ = units.parse_unit(input_string)
        if value is None:
            return None

        return units.convert_unit(input_string, "")

    def vb_scale(self, viewbox, p_a_r, doc_width, doc_height):
        """
        Parse SVG viewbox and generate scaling parameters.
        Reference documentation: https://www.w3.org/TR/SVG11/coords.html

        Inputs:
            viewbox:         Contents of SVG viewbox attribute
            p_a_r:      Contents of SVG preserveAspectRatio attribute
            doc_width:  Width of SVG document
            doc_height: Height of SVG document

        Output: s_x, s_y, o_x, o_y
            Scale parameters (s_x, s_y) and offset parameters (o_x, o_y)

        """
        if viewbox is None:
            return 1, 1, 0, 0  # No viewbox; return default transform
        vb_array = viewbox.strip().replace(", ", " ").split()

        if len(vb_array) < 4:
            return 1, 1, 0, 0  # invalid viewbox; return default transform

        min_x = float(vb_array[0])  # Viewbox offset: x
        min_y = float(vb_array[1])  # Viewbox offset: y
        width = float(vb_array[2])  # Viewbox width
        height = float(vb_array[3])  # Viewbox height

        if width <= 0 or height <= 0:
            return 1, 1, 0, 0  # invalid viewbox; return default transform

        if doc_width is None or doc_height is None:
            raise AbortExtension(
                _("Width or height attribute missing on toplevel <svg> tag")
            )

        d_width = float(doc_width)
        d_height = float(doc_height)

        if d_width <= 0 or d_height <= 0:
            return 1, 1, 0, 0  # invalid document size; return default transform

        ar_doc = d_height / d_width  # Document aspect ratio
        ar_vb = height / width  # Viewbox aspect ratio

        # Default values of the two preserveAspectRatio parameters:
        par_align = "xmidymid"  # "align" parameter(lowercased)
        par_mos = "meet"  # "meetOrSlice" parameter

        if p_a_r is not None:
            par_array = p_a_r.strip().replace(", ", " ").lower().split()
            if len(par_array) > 0:
                par0 = par_array[0]
                if par0 == "defer":
                    if len(par_array) > 1:
                        par_align = par_array[1]
                        if len(par_array) > 2:
                            par_mos = par_array[2]
                else:
                    par_align = par0
                    if len(par_array) > 1:
                        par_mos = par_array[1]

        if par_align == "none":
            # Scale document to fill page. Do not preserve aspect ratio.
            # This is not default behavior, nor what happens if par_align
            # is not given; the "none" value must be _explicitly_ specified.

            s_x = d_width / width
            s_y = d_height / height
            o_x = -min_x
            o_y = -min_y
            return s_x, s_y, o_x, o_y

        """
        Other than "none", all situations fall into two classes:

        1)  (ar_doc >= ar_vb AND par_mos == "meet")
               or (ar_doc < ar_vb AND par_mos == "slice")
            -> In these cases, scale document up until VB fills doc in X.

        2)   All other cases, i.e.,
           (ar_doc < ar_vb AND par_mos == "meet")
               or (ar_doc >= ar_vb AND par_mos == "slice")
            -> In these cases, scale document up until VB fills doc in Y.

        Note in cases where the scaled viewbox exceeds the document
        (page) boundaries (all "slice" cases and many "meet" cases where
        an offset value is given) that this routine does not perform
        any clipping, but subsequent clipping to the page boundary
        is appropriate.

        Besides "none", there are 9 possible values of par_align:
            xminymin xmidymin xmaxymin
            xminymid xmidymid xmaxymid
            xminymax xmidymax xmaxymax
        """

        if ((ar_doc >= ar_vb) and (par_mos == "meet")) or (
            (ar_doc < ar_vb) and (par_mos == "slice")
        ):
            # Case 1: Scale document up until VB fills doc in X.

            s_x = d_width / width
            s_y = s_x  # Uniform aspect ratio
            o_x = -min_x

            scaled_vb_height = ar_doc * width
            excess_height = scaled_vb_height - height

            if par_align in {"xminymin", "xmidymin", "xmaxymin"}:
                # Case: Y-Min: Align viewbox to minimum Y of the viewport.
                o_y = -min_y
                # OK: tested with Tall-Meet, Wide-Slice

            elif par_align in {"xminymax", "xmidymax", "xmaxymax"}:
                # Case: Y-Max: Align viewbox to maximum Y of the viewport.
                o_y = -min_y + excess_height
                #  OK: tested with Tall-Meet, Wide-Slice

            else:  # par_align in {"xminymid", "xmidymid", "xmaxymid"}:
                # Default case: Y-Mid: Center viewbox on page in Y
                o_y = -min_y + excess_height / 2
                # OK: Tested with Tall-Meet, Wide-Slice

            return s_x, s_y, o_x, o_y

        # Case 2: Scale document up until VB fills doc in Y.

        s_y = d_height / height
        s_x = s_y  # Uniform aspect ratio
        o_y = -min_y

        scaled_vb_width = height / ar_doc
        excess_width = scaled_vb_width - width

        if par_align in {"xminymin", "xminymid", "xminymax"}:
            # Case: X-Min: Align viewbox to minimum X of the viewport.
            o_x = -min_x
            # OK: Tested with Tall-Slice, Wide-Meet

        elif par_align in {"xmaxymin", "xmaxymid", "xmaxymax"}:
            # Case: X-Max: Align viewbox to maximum X of the viewport.
            o_x = -min_x + excess_width
            # Need test: Tall-Slice, Wide-Meet

        else:  # par_align in {"xmidymin", "xmidymid", "xmidymax"}:
            # Default case: X-Mid: Center viewbox on page in X
            o_x = -min_x + excess_width / 2
            # OK: Tested with Tall-Slice, Wide-Meet

        return s_x, s_y, o_x, o_y

    def strip_quotes(self, fontname):
        """
        A multi-word font name may have a leading and trailing
        single or double quotes, depending on the source.
        If so, remove those quotes.
        """

        if fontname.startswith("'") and fontname.endswith("'"):
            return fontname[1:-1]
        if fontname.startswith('"') and fontname.endswith('"'):
            return fontname[1:-1]
        return fontname

    def parse_svg_font(self, node_list):
        """
         Parse an input svg, searching for an SVG font. If an
         SVG font is found, parse it and return a "digest" containing
         structured information from the font. See below for more
         about the digest format.

         If the font is not found cannot be parsed, return none.

         Notable limitations:

        (1) This function only parses the first font face found within the
         tree. We may, in the future, support discovering multiple fonts
         within an SVG file.

        (2) We are only processing left-to-right and horizontal text,
         not vertical text nor RTL.

        (3) This function currently performs only certain recursive searches,
         within the <defs> element. It will not discover fonts nested within
         groups or other elements. So far as we know, that is not a limitation
         in practice. (If you have a counterexample please contact Evil Mad
         Scientist tech support and let us know!)

        (4) Kerning details are not implemented yet.
        """

        digest = None

        if node_list is None:
            return None

        for node in node_list:
            if isinstance(node, Defs):
                return self.parse_svg_font(node)  # Recursive call

            if isinstance(node, SVGfont):
                """
                === Internal structure for storing font information ===

                We parse the SVG font file and create a keyed "digest"
                from it that we can use while rendering text on the page.

                This "digest" will be added to a dictionary that maps
                each font family name to a single digest.

                The digest itself is a dictionary with the following
                keys, some of which may have empty values. This format
                will allow us to add additional keys at a later date,
                to support additional SVG font features.

                    font_id (a string)

                    font_family (a string)

                    glyphs
                        A dictionary mapping unicode points to a specific
                        dictionary for each point.  See below for more about
                        the key format.
                        The dictionary for a given point will include keys:
                            glyph_name(string)
                            horiz_adv_x(numeric)
                            d(string)

                    missing_glyph
                        A dictionary for a single code point, with keys:
                            horiz_adv_x(numeric)
                            d(string)

                    geometry
                        A dictionary containing geometric data
                        Keys will include:
                            horiz_adv_x(numeric) -- Default value
                            units_per_em(numeric)
                            ascent(numeric)
                            descent(numeric)
                            x_height(numeric)
                            cap_height(numeric)
                            bbox (string)
                            underline_position(numeric)
                    scale
                        A numeric scaling factor computed from the
                        units_per_em value, which gives the overall scale
                """

                digest = dict()
                geometry = dict()
                glyphs = dict()
                missing_glyph = dict()

                digest["font_id"] = node.get("id")

                horiz_adv_x = node.get("horiz-adv-x")

                if horiz_adv_x is not None:
                    geometry["horiz_adv_x"] = float(horiz_adv_x)
                # Note: case of no horiz_adv_x value is not handled.

                for element in node:
                    if isinstance(element, Glyph):
                        # First, because it is the most common element
                        try:
                            uni_text = element.get("unicode")
                        except:
                            # Can't use this point if no unicode mapping.
                            continue

                        if uni_text is None:
                            continue

                        if uni_text in glyphs:
                            # Skip if that unicode point is already in the
                            # list of glyphs.(There is not currently support
                            # for alternate glyphs in the font.)
                            continue

                        glyph_dict = dict()
                        glyph_dict["glyph_name"] = element.get("glyph-name")

                        horiz_adv_x = element.get("horiz-adv-x")

                        if horiz_adv_x is not None:
                            glyph_dict["horiz_adv_x"] = float(horiz_adv_x)
                        else:
                            glyph_dict["horiz_adv_x"] = geometry["horiz_adv_x"]

                        glyph_dict["d"] = element.get("d")  # SVG path data
                        glyphs[uni_text] = glyph_dict

                    elif isinstance(element, FontFace):
                        digest["font_family"] = element.get("font-family")
                        units_per_em = element.get("units-per-em")

                        if units_per_em is None:
                            # Default: 1000, per SVG specification.
                            geometry["units_per_em"] = 1000.0
                        else:
                            geometry["units_per_em"] = float(units_per_em)

                        ascent = element.get("ascent")
                        if ascent is not None:
                            geometry["ascent"] = float(ascent)

                        descent = element.get("descent")
                        if descent is not None:
                            geometry["descent"] = float(descent)

                        """
                        # Skip these attributes that we are not currently using
                        geometry['x_height'] = element.get('x-height')
                        geometry['cap_height'] = element.get('cap-height')
                        geometry['bbox'] = element.get('bbox')
                        geometry['underline_position'] = element.get('underline-position')
                        """

                    elif isinstance(element, MissingGlyph):
                        horiz_adv_x = element.get("horiz-adv-x")

                        if horiz_adv_x is not None:
                            missing_glyph["horiz_adv_x"] = float(horiz_adv_x)
                        else:
                            missing_glyph["horiz_adv_x"] = geometry["horiz_adv_x"]

                        missing_glyph["d"] = element.get("d")  # SVG path data
                        digest["missing_glyph"] = missing_glyph

                # Main scaling factor
                digest["scale"] = 1.0 / geometry["units_per_em"]

                digest["glyphs"] = glyphs
                digest["geometry"] = geometry

                return digest
        return None

    def load_font(self, fontname):
        """
        Attempt to load an SVG font from a file in our list
        of (likely) SVG font files.
        If we can, add the contents to the font library.
        Otherwise, add a "None" entry to the font library.
        """

        if fontname is None:
            return

        if fontname in self.font_dict:
            return  # Awesome: The font is already loaded.

        if fontname in self.font_file_list:
            the_path = self.font_file_list[fontname]
        else:
            self.font_dict[fontname] = None
            return  # Font not located.
        try:
            """
            Check to see if there is an SVG font file for us to read.

            At present, only one font file will be read per font family;
            the name of the file must be FONT_NAME.svg, where FONT_NAME
            is the name of the font family.

            Only the first font found in the font file will be read.
            Multiple weights and styles within a font family are not
            presently supported.
            """
            font_svg = load_svg(the_path)
            self.font_dict[fontname] = self.parse_svg_font(font_svg.getroot())

        except IOError:
            self.font_dict[fontname] = None
        except:
            inkex.errormsg(_("Error parsing SVG font at {}").format(str(the_path)))
            self.font_dict[fontname] = None

    def font_table(self):
        """
        Generate display table of all available SVG fonts
        """

        self.options.preserve_text = False

        # Embed text in group to make manipulation easier:
        group = self.svg.get_current_layer().add(Group())
        for fontname in self.font_file_list:
            self.load_font(fontname)

        font_size = 0.2  # in inches -- will be scaled by viewbox factor.
        font_size_text = str(font_size / self.vb_scale_factor) + "px"

        labeltext_style = Style(
            {
                "stroke": "none",
                "font-size": font_size_text,
                "fill": "black",
                "font-family": "sans-serif",
                "text-anchor": "end",
            }
        )

        x_offset = font_size / self.vb_scale_factor
        y_offset = 1.5 * x_offset
        y = y_offset

        for fontname in sorted(self.font_dict):
            if self.font_dict[fontname] is None:
                continue  # If the SVG file did NOT contain a font, skip it.

            text_attribs = {"x": "0", "y": str(y), "hershey-ignore": "true"}
            textline = group.add(TextElement(**text_attribs))
            textline.text = fontname
            textline.style = labeltext_style
            text_attribs = {"x": str(x_offset), "y": str(y)}

            sampletext_style = Style(
                {
                    "stroke": "none",
                    "font-size": font_size_text,
                    "fill": "black",
                    "font-family": fontname,
                    "text-anchor": "start",
                }
            )
            sampleline = group.add(TextElement(**text_attribs))

            try:  # python 2
                sampleline.text = self.options.sample_text.decode("utf-8")
            except AttributeError:  # python 3
                sampleline.text = self.options.sample_text

            sampleline.style = sampletext_style
            y += y_offset
        self.recursively_traverse_svg(group, self.doc_transform)

    def glyph_table(self):
        """
        Generate display table of glyphs within the current SVG font. Sorted display of
        all printable characters in the font _except_ missing glyph.
        """

        self.options.preserve_text = False

        fontname = self.font_load_wrapper("not_a_font_name")  # force load of default

        if self.font_load_fail:
            inkex.errormsg(_("Font not found; Unable to generate glyph table."))
            return

        # Embed in group to make manipulation easier:
        group = self.svg.get_current_layer().add(Group())

        # missing_glyph = self.font_dict[fontname]['missing_glyph']

        glyph_count = 0
        for glyph in self.font_dict[fontname]["glyphs"]:
            if self.font_dict[fontname]["glyphs"][glyph]["d"] is not None:
                glyph_count += 1

        columns = int(math.floor(math.sqrt(glyph_count)))

        font_size = 0.4  # in inches -- will be scaled by viewbox factor.
        font_size_text = str(font_size / self.vb_scale_factor) + "px"

        glyph_style = Style(
            {
                "stroke": "none",
                "font-size": font_size_text,
                "fill": "black",
                "font-family": fontname,
                "text-anchor": "start",
            }
        )

        x_offset = 1.5 * font_size / self.vb_scale_factor
        y_offset = x_offset
        x = x_offset
        y = y_offset

        draw_position = 0

        for glyph in sorted(self.font_dict[fontname]["glyphs"]):
            if self.font_dict[fontname]["glyphs"][glyph]["d"] is None:
                continue
            y_pos, x_pos = divmod(draw_position, columns)
            x = x_offset * (x_pos + 1)
            y = y_offset * (y_pos + 1)
            text_attribs = {"x": str(x), "y": str(y)}
            sampleline = group.add(TextElement(**text_attribs))
            sampleline.text = glyph
            sampleline.style = glyph_style
            draw_position = draw_position + 1

        self.recursively_traverse_svg(group, self.doc_transform)

    def find_font_files(self):
        """
         Create list of "plausible" SVG font files

         List items in primary svg_fonts directory, typically located in the
         directory where this script is being executed from.

         If there is text given in the "Other name/path" input, that text may
         represent one of the following:

        (A) The name of a font file, located in the svg_fonts directory.
             - This may be given with or without the .svg suffix.
             - If it is a font file, and the font face selected is "other",
                 then use this as the default font face.

        (B) The path to a font file, located elsewhere.
             - If it is a font file, and the font face selected is "other",
                 then use this as the default font face.
             - ALSO: Search the directory where that file is located for
                 any other SVG fonts.

        (C) The path to a directory
             - It may or may not have a trailing separator
             - Search that directory for SVG fonts.

         This function will create a list of available files that
         appear to be SVG(SVG font) files. It does not parse the files.
         We will format it as a dictionary, that maps each file name
         (without extension) to a path.
        """

        self.font_file_list = dict()

        # List contents of primary font directory:
        font_directory_name = "svg_fonts"

        font_dir = os.path.realpath(os.path.join(os.getcwd(), font_directory_name))
        for dir_item in os.listdir(font_dir):
            if dir_item.endswith((".svg", ".SVG")):
                file_path = os.path.join(font_dir, dir_item)
                if os.path.isfile(file_path):  # i.e., if not a directory
                    root, _ = os.path.splitext(dir_item)
                    self.font_file_list[root] = file_path

        # split off file extension(e.g., ".svg")
        root, _ = os.path.splitext(self.options.otherfont)

        # Check for case "(A)": Input text is the name
        # of an item in the primary font directory.
        if root in self.font_file_list:
            # If we already have that name in our font_file_list,
            # and "other" is selected, this is now
            # our default font face.
            if self.options.fontface == "other":
                self.options.fontface = root
            return

        test_path = os.path.realpath(self.options.otherfont)

        # Check for case "(B)": A file, not in primary font directory
        if os.path.isfile(test_path):
            directory, file_name = os.path.split(test_path)
            root, ext = os.path.splitext(file_name)
            self.font_file_list[root] = test_path

            if self.options.fontface == "other":
                self.options.fontface = root

            # Also search the directory where that file
            # was located for other SVG files(which may be fonts)

            for dir_item in os.listdir(directory):
                if dir_item.endswith((".svg", ".SVG")):
                    file_path = os.path.join(directory, dir_item)
                    if os.path.isfile(file_path):  # i.e., if not a directory
                        root, _ = os.path.splitext(dir_item)
                        self.font_file_list[root] = file_path
            return

        # Check for case "(C)": A directory name
        if os.path.isdir(test_path):
            for dir_item in os.listdir(test_path):
                if dir_item.endswith((".svg", ".SVG")):
                    file_path = os.path.join(test_path, dir_item)
                    if os.path.isfile(file_path):  # i.e., if not a directory
                        root, _ = os.path.splitext(dir_item)
                        self.font_file_list[root] = file_path

    def font_load_wrapper(self, fontname):
        """

        This implements the following logic:

        * Check to see if the font name is in our lookup table of fonts,
            self.font_dict

        * If the font is not listed in font_dict[]:
            * Check to see if there is a corresponding SVG font file that
            can be opened and parsed.

            * If the font can be opened and parsed:
                * Add that font to font_dict.
            * Otherwise
                * Add the font name to font_dict as None.

        * If the font has value None in font_dict:
            * Try to load fallback font.

        * Fallback font:
            * If an SVG font matching that in the SVG is not available,
            check to see if the default font is available. That font
            is given by self.options.fontface

        * If a font is loaded and available, return the font name.
            Otherwise, return none.

        """

        self.load_font(fontname)  # Load the font if available

        """
        It *may* be worth building one stroke font (e.g., Hershey Sans 1-stroke) as a
            variable defined in this file so that it can be used even if no external
            SVG font files are available.
        """

        if self.font_dict[fontname] is None:
            # If we were not able to load the requested font::
            fontname = self.options.fontface  # Fallback
            if fontname not in self.font_dict:
                self.load_font(fontname)
            else:
                pass

        if self.font_dict[fontname] is None:
            self.font_load_fail = (
                True  # Set a flag so that we only generate one copy of this error.
            )
            return None
        return fontname

    def get_font_char(self, fontname, char):
        """
        Given a font face name and a character(unicode point),
            return an SVG path, horizontal advance value,
            and scaling factor.

        If the font is not available by name, use the default font.
        """

        fontname = self.font_load_wrapper(fontname)  # Load the font if available

        if fontname is None:
            return None

        try:
            scale_factor = self.font_dict[fontname]["scale"]
        except:
            scale_factor = 0.001  # Default: 1/1000

        try:
            if char not in self.font_dict[fontname]["glyphs"]:
                x_adv = self.font_dict[fontname]["missing_glyph"]["horiz_adv_x"]

                return (
                    self.font_dict[fontname]["missing_glyph"]["d"],
                    x_adv,
                    scale_factor,
                )
            x_adv = self.font_dict[fontname]["glyphs"][char]["horiz_adv_x"]

            return self.font_dict[fontname]["glyphs"][char]["d"], x_adv, scale_factor
        except:
            return None

    def handle_viewbox(self):
        """
        Wrapper function for processing viewbox information
        """

        self.svg_height = self.getlength_inch("height")
        self.svg_width = self.getlength_inch("width")

        self.svg = self.document.getroot()
        viewbox = self.svg.get("viewBox")
        if viewbox:
            p_a_r = self.svg.get("preserveAspectRatio")
            s_x, s_y, o_x, o_y = self.vb_scale(
                viewbox, p_a_r, self.svg_width, self.svg_height
            )
        else:
            s_x = 1.0 / float(self.PX_PER_INCH)  # Handle case of no viewbox
            s_y = s_x
            o_x = 0.0
            o_y = 0.0

        # Initial transform of document is based on viewbox, if present:
        self.doc_transform = Transform(scale=(s_x, s_y), translate=(o_x, o_y))

        self.vb_scale_factor = (s_x + s_y) / 2.0
        # In case of non-square aspect ratio, use average value.

    def draw_svg_text(self, chardata, parent):
        """
        Render an individual svg glyph
        """
        char = chardata["char"]
        font_family = chardata["font_family"]
        offset = chardata["offset"]
        vertoffset = chardata["vertoffset"]
        font_height = chardata["font_height"]
        font_scale = 1.0

        # Stroke scale factor, including external transformations:
        stroke_scale = chardata["stroke_scale"] * self.vb_scale_factor

        try:
            path_string, adv_x, scale_factor = self.get_font_char(font_family, char)
        except:
            adv_x = 0
            path_string = None
            scale_factor = 1.0

        if self.font_load_fail:
            return 0

        font_scale *= scale_factor * font_height

        h_offset = 0
        v_offset = 0

        # SVG fonts use inverted Y axis; mirror vertically
        scale_transform = Transform(scale=(font_scale, -font_scale))

        # Combine scales of external transformations with the scaling
        # applied by this function:
        _scale = font_scale * stroke_scale
        if _scale == 0:
            _scale = 1
        stroke_width = self.render_width / _scale

        # Stroke-width is a css style element; cannot use scientific notation.
        # Thus, use variable width for encoding the stroke width factor:

        log_ten = math.log10(stroke_width)
        if log_ten > 0:  # For stroke_width > 1
            width_string = "{0:.3f}in".format(stroke_width)
        else:
            prec = int(math.ceil(-log_ten) + 3)
            width_string = "{0:.{1}f}in".format(stroke_width, prec)

        p_style = {"stroke-width": width_string}

        the_transform = Transform(translate=(offset + h_offset, vertoffset + v_offset))
        the_transform @= scale_transform

        if path_string is not None:
            path_element = parent.add(PathElement())
            path_element.set_path(path_string)
            path_element.style = p_style
            path_element.transform = the_transform
            self.output_generated = True

        return offset + float(adv_x) * font_scale  # new horizontal offset value

    def recursive_get_encl_transform(self, node):
        """
        Determine the cumulative transform which node inherits from
        its chain of ancestors.
        """
        node = node.getparent()
        if node is not None:
            parent_transform = self.recursive_get_encl_transform(node)
            node_transform = node.get("transform", None)
            if node_transform is None:
                return parent_transform
            trans = Transform(node_transform).matrix

            if parent_transform is None:
                return trans
            return Transform(parent_transform) @ Transform(trans)
        return self.doc_transform

    def recursively_parse_flowroot(self, node_list, parent_info):
        """
        Parse a flowroot node and its children
        """

        # By default, inherit these values from parent:
        font_height_local = parent_info["font_height"]
        font_family_local = parent_info["font_family"]
        line_spacing_local = parent_info["line_spacing"]
        text_align_local = parent_info["align"]

        for node in node_list:
            node_style = node.style

            font_height = node_style("font-size")
            try:
                font_height_local = self.units_to_userunits(font_height)
            except TypeError:
                pass

            font_family_local = self.strip_quotes(node_style("font-family"))

            try:
                line_spacing = node_style("line-height")
                if "%" in line_spacing:  # Handle percentage line spacing(e.g., 125%)
                    line_spacing_local = float(line_spacing.rstrip("%")) / 100.0
                elif line_spacing == "normal":
                    line_spacing_local = 1.25  # Inkscape default line spacing
                else:
                    line_spacing_local = self.units_to_userunits(line_spacing)
            except TypeError:
                pass

            text_align_local = node_style("text-align")  # Use text-anchor in text nodes

            if node.text is not None:
                self.text_string += node.text

                for _ in node.text:
                    self.text_families.append(font_family_local)
                    self.text_heights.append(font_height_local)
                    self.text_spacings.append(line_spacing_local)
                    self.text_aligns.append(text_align_local)

            if isinstance(node, (FlowPara, FlowSpan)):
                the_style = dict()
                the_style["font_height"] = font_height_local
                the_style["font_family"] = font_family_local
                the_style["line_spacing"] = line_spacing_local
                the_style["align"] = text_align_local

                self.recursively_parse_flowroot(node, the_style)

            if node.tail is not None:
                # By default, inherit these values from parent:
                font_height_local = parent_info["font_height"]
                font_family_local = parent_info["font_family"]
                line_spacing_local = parent_info["line_spacing"]

                text_align_local = parent_info["align"]
                self.text_string += node.tail
                for _ in node.tail:
                    self.text_families.append(font_family_local)
                    self.text_heights.append(font_height_local)
                    self.text_spacings.append(line_spacing_local)
                    self.text_aligns.append(text_align_local)

            if isinstance(node, FlowPara):
                self.text_string += "\n"  # Conclude every flowpara with a return
                self.text_families.append(font_family_local)
                self.text_heights.append(font_height_local)
                self.text_spacings.append(line_spacing_local)
                self.text_aligns.append(text_align_local)

    def recursively_parse_text(self, node, parent_info):
        """
        parse a text node and its children
        """

        # By default, inherit these values from parent:
        font_height_local = parent_info["font_height"]
        font_family_local = parent_info["font_family"]
        anchor_local = parent_info["anchor"]
        x_local = parent_info["x_pos"]
        y_local = parent_info["y_pos"]
        parent_line_spacing = parent_info["line_spacing"]

        node_style = node.style

        font_height = node_style("font-size")
        try:
            font_height_local = self.units_to_userunits(font_height)
        except TypeError:
            pass

        font_family_local = self.strip_quotes(node_style("font-family"))
        anchor_local = node_style("text-anchor")  # Use text-anchor in text nodes

        try:
            x_temp = node.get("x")
            if x_temp is not None:
                x_local = x_temp
        except ValueError:
            pass

        try:
            y_temp = node.get("y")
            if y_temp is not None:
                y_local = y_temp
            else:
                # Special case, to handle multi-line text given by tspan
                # elements that do not have y values
                if y_local is None:
                    y_local = 0
                y_local = (
                    float(y_local)
                    + self.line_number * parent_line_spacing * font_height_local
                )
        except ValueError:
            pass

        if node.text is not None:
            self.text_string += node.text

            for _ in node.text:
                self.text_families.append(font_family_local)
                self.text_heights.append(font_height_local)
                self.text_aligns.append(anchor_local)
                self.text_x.append(x_local)
                self.text_y.append(y_local)

        for sub_node in node:
            # If text is located within a sub_node of this node,
            #   process that sub_node, with this very routine.

            if isinstance(sub_node, Tspan):
                # Note: There may be additional types of text tags that
                #   we should recursively search as well.
                node_info = dict()
                node_info["font_height"] = font_height_local
                node_info["font_family"] = font_family_local
                node_info["anchor"] = anchor_local
                node_info["x_pos"] = x_local
                node_info["y_pos"] = y_local
                node_info["line_spacing"] = parent_line_spacing

                adv_line = False
                role = sub_node.get("sodipodi:role")
                if role == "line":
                    adv_line = True

                self.recursively_parse_text(sub_node, node_info)

                # Increment line after tspan if it is labeled as a line
                if adv_line:
                    self.line_number = self.line_number + 1

        if node.tail is not None:
            _stripped_tail = node.tail.strip()
            if _stripped_tail is not None:
                # By default, inherit these values from parent:
                font_height_local = parent_info["font_height"]
                font_family_local = parent_info["font_family"]
                text_align_local = parent_info["anchor"]
                x_local = parent_info["x_pos"]
                y_local = parent_info["y_pos"]
                self.text_string += _stripped_tail
                for _ in _stripped_tail:
                    self.text_heights.append(font_height_local)
                    self.text_families.append(font_family_local)
                    self.text_aligns.append(text_align_local)
                    self.text_x.append(x_local)
                    self.text_y.append(y_local)

    def recursively_traverse_svg(
        self,
        anode_list,
        mat_current=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
        parent_visibility="visible",
    ):
        """
        recursively parse the full document and its children,
        looking for nodes that may contain text
        """

        for node in anode_list:
            # Ignore invisible nodes
            vis = node.get("visibility", parent_visibility)
            if vis == "inherit":
                vis = parent_visibility
            if vis in ("hidden", "collapse"):
                continue

            # First apply the current matrix transform to this node's tranform
            _matrix = node.transform
            mat_new = Transform(mat_current) @ Transform(_matrix)

            if isinstance(node, Group):
                recurse_group = True
                ink_label = node.get("inkscape:label")

                if not ink_label:
                    pass
                else:
                    if ink_label == "Hershey Text":
                        recurse_group = (
                            False  # Do not traverse groups of rendered text.
                        )
                if recurse_group:
                    self.recursively_traverse_svg(node, mat_new, vis)

            elif isinstance(node, Use):
                # A <use> element refers to another SVG element via an xlink:href="#blah"
                # attribute.  We will handle the element by doing an XPath search through
                # the document, looking for the element with the matching id="blah"
                # attribute.  We then recursively process that element after applying
                # any necessary(x, y) translation.
                #
                # Notes:
                #  1. We ignore the height and width attributes as they do not apply to
                #     path-like elements, and
                #  2. Even if the use element has visibility="hidden", SVG still calls
                #     for processing the referenced element.  The referenced element is
                #     hidden only if its visibility is "inherit" or "hidden".

                refnode = node.href
                if refnode is None:
                    continue  # missing reference

                local_transform = Transform(_matrix)
                x = float(node.get("x", "0"))
                y = float(node.get("y", "0"))
                # Note: the transform has already been applied
                if (x != 0) or (y != 0):
                    _trans_string = "translate({0:.6E}, {1:.6E})".format(x, y)
                    ref_transform = Transform(_matrix) @ Transform(_trans_string)
                else:
                    ref_transform = local_transform

                try:
                    ref_group = anode_list.add(Group())  # Add a subgroup
                except AttributeError:
                    inkex.errormsg(
                        _("Unable to process text. Consider unlinking cloned text.")
                    )
                    continue

                # Tests are not using the preset seed for this atm
                # if 'id' not in ref_group.attrib:
                #    ref_group.set_random_id('')

                ref_group.set("transform", ref_transform)

                ref_group.append(deepcopy(refnode))

                for sub_node in ref_group:
                    # The copied text elements should be removed at the end,
                    # or they will persist if original elements are preserved.
                    self.nodes_to_delete.append(sub_node)

                # Preserve original element?
                if not self.options.preserve_text:
                    self.nodes_to_delete.append(node)

            elif isinstance(node, (TextElement, FlowRoot)):
                # Flag for when we start a new line of text, for use with indents:
                self.new_line = True

                start_x = (
                    0  # Defaults; Fail gracefully in case xy position is not given.
                )
                start_y = 0

                # Default line spacing and font height: 125%, 16 px
                line_spacing = self.units_to_userunits("1.25")
                font_height = self.units_to_userunits("16px")

                start_x = node.get("x")  # XY Position of element
                start_y = node.get("y")

                bounding_rect = False
                # rect_height = 100        #default size of bounding rectangle for flowroot object
                rect_width = (
                    100  # default size of bounding rectangle for flowroot object
                )
                transform = ""  # transform(scale, translate, matrix, etc.)
                text_align = "start"

                try:
                    hershey_ignore = node.get("hershey-ignore")
                    if hershey_ignore is not None:
                        continue  # If the attribute is present, skip this node.
                except ValueError:
                    pass

                node_style = node.style

                try:
                    font_height_temp = node_style.get("font-size", 16)
                    font_height = self.units_to_userunits(font_height_temp)
                except TypeError:
                    pass

                font_family = self.strip_quotes(node_style("font-family"))

                try:
                    line_spacing_temp = node_style("line-height")
                    if (
                        "%" in line_spacing_temp
                    ):  # Handle percentage line spacing(e.g., 125%)
                        line_spacing = float(line_spacing_temp.rstrip("%")) / 100.0
                    elif line_spacing_temp == "normal":
                        line_spacing = 1.25  # Inkscape default line spacing
                    else:
                        line_spacing = self.units_to_userunits(line_spacing_temp)
                except TypeError:
                    pass

                try:
                    transform = node.transform
                except ValueError:
                    pass

                if transform is not None:
                    transform2 = Transform(transform).matrix

                    """
                    Compute estimate of transformation scale applied to
                    this element, for purposes of calculating the
                    stroke width to apply. When all transforms are applied
                    and our elements are displayed on the page, we want the
                    final visible stroke width to be reasonable.
                    Transformation matrix is [[a c e][b d f]]
                    scale_x = sqrt(a * a + b * b),
                    scale_y = sqrt(c * c + d * d)
                    Take estimated scale as the mean of the two.
                    """

                    scale_x = math.sqrt(
                        transform2[0][0] * transform2[0][0]
                        + transform2[1][0] * transform2[1][0]
                    )
                    scale_y = math.sqrt(
                        transform2[0][1] * transform2[0][1]
                        + transform2[1][1] * transform2[1][1]
                    )

                    scale_r = (scale_x + scale_y) / 2.0  # Average. ¯\_(ツ)_/¯
                else:
                    scale_r = 1.0

                the_id = node.get("id")

                # Initialize text attribute lists for each top-level text object:
                self.text_string = ""
                self.text_families = []  # Lis of font family for characters in the string
                self.text_heights = []  # List of font heights
                self.text_spacings = []  # List of vertical line heights
                self.text_aligns = []  # List of horizontal alignment values
                self.text_x = []  # List; x-coordinate of text line start
                self.text_y = []  # List; y-coordinate of text line start

                # Group generated paths together, to make the rendered letters
                # easier to manipulate in Inkscape once generated:
                parent = node.getparent()

                group = parent.add(Group())
                group.label = "Hershey Text"

                style = {
                    "stroke": "#000000",
                    "fill": "none",
                    "stroke-linecap": "round",
                    "stroke-linejoin": "round",
                }

                # Apply rounding to ends to improve final engraved text appearance.
                group.style = style
                # Some common variables used in both cases A and B:
                str_pos = 0  # Position through the full string that we are rendering
                i = 0  # Dummy(index) variable for looping over letters in string
                w = 0  # Initial spacing offset
                w_temp = 0  # Temporary variable for horizontal spacing offset
                width_this_line = (
                    0  # Estimated width of characters to be stored on this line
                )

                """
                CASE A: Handle flowed text nodes
                """

                if isinstance(node, FlowRoot):
                    try:
                        text_align = node_style["text-align"]
                        # Use text-align, not text-anchor, in flowroot
                    except KeyError:
                        pass

                    # selects the flowRegion's child(svg:rect) to get @X and @Y
                    flowref = self.svg.getElement(
                        '/svg:svg//*[@id="%s"]/svg:flowRegion[1]' % the_id
                    )[0]

                    if isinstance(flowref, Rectangle):
                        start_x = flowref.left
                        start_y = flowref.top
                        rect_height = flowref.height
                        rect_width = flowref.width
                        bounding_rect = True

                    elif isinstance(flowref, Use):
                        pass

                        # A <use> element refers to another SVG element via an xlink:href="#blah"
                        # attribute.  We will handle the element by doing an XPath search through
                        # the document, looking for the element with the matching id="blah"
                        # attribute.  We then recursively process that element after applying
                        # any necessary(x, y) translation.
                        #
                        # Notes:
                        #  1. We ignore the height and width attributes as they do not apply to
                        #     path-like elements, and
                        #  2. Even if the use element has visibility="hidden", SVG still calls
                        #     for processing the referenced element.  The referenced element is
                        #     hidden only if its visibility is "inherit" or "hidden".
                        #  3. We may be able to unlink clones using the code in pathmodifier.py

                        # The following code can render text flowed into a rectangle object.
                        # HOWEVER, it does not handle the various transformations that could occur
                        # be present on those objects, and does not handle more general cases, such
                        # as a rotated rectangle -- for which text *should* flow in a diamond shape.
                        # For the time being, we skip these and issue a warning.
                        #
                        # refid = flowref.get('xlink:href')
                        # if refid is not None:
                        #     # [1:] to ignore leading '#' in reference
                        #     path = '//*[@id="%s"]' % refid[1:]
                        #     refnode = flowref.xpath(path)
                        #     if refnode is not None:
                        #         refnode = refnode[0]
                        #         if isinstance(refnode, Rectangle):
                        #             start_x = refnode.get('x")
                        #             start_y = refnode.get('y")
                        #             rect_height = refnode.get('height")
                        #             rect_width = refnode.get('width")
                        #             bounding_rect = True

                    if not bounding_rect:
                        self.warn_unflow = True
                        continue

                    """
                    Recursively loop through content of the flowroot object,
                    looping through text, flowpara, and other things.

                    Create multiple lists: One of text content,
                    others of style that should be applied to that content.

                    then, loop through those lists, one line at a time,
                    finding how many words fit on a line, etc.
                    """

                    the_style = dict()
                    the_style["font_height"] = font_height
                    the_style["font_family"] = font_family
                    the_style["line_spacing"] = line_spacing
                    the_style["align"] = text_align

                    self.recursively_parse_flowroot(node, the_style)

                    if self.text_string == "":
                        continue  # No convertable text in this SVG element.

                    if self.text_string.isspace():
                        continue  # No convertable text in this SVG element.

                    # Initial vertical offset for the flowed text block:
                    v = 0

                    # Record that we are on the first line of the paragraph
                    # for setting the v position of the first line.
                    first_line = True

                    # Keep track of text height on first line, for moving entire text box:
                    y_offs_overall = 0

                    # Split text by lines AND make a list of how long each
                    # line is, including the newline characters.
                    # We need to keep track of this to match up styling
                    # information to the printable characters.

                    text_lines = self.text_string.splitlines()
                    extd_text_lines = self.text_string.splitlines(True)
                    str_pos_eol = 0  # str_pos after end of previous text_line.

                    nbsp = "\xa0"  # Unicode non-breaking space character

                    for line_number, text_line in enumerate(text_lines):
                        line_length = len(text_line)
                        extd_line_length = len(extd_text_lines[line_number])

                        i = 0  # Position within this text_line.

                        # A given text_line may take more than one strip
                        # to render, if it overflows our box width.

                        line_start = 0  # Value of i when the current strip started.

                        if line_length == 0:
                            str_pos_temp = str_pos_eol
                            char_height = float(self.text_heights[str_pos_temp])
                            charline_spacing = float(self.text_spacings[str_pos_temp])
                            char_v_spacing = charline_spacing * char_height
                            v = v + char_v_spacing
                        else:
                            while i < line_length:
                                word_start = (
                                    i  # Value of i at beginning of the current word.
                                )

                                while i < line_length:  # Step through the line
                                    # until we reach the end of the line or word.
                                    # (i.e., until we reach whitespace)
                                    character = text_line[
                                        i
                                    ]  # character is unicode(not byte string)
                                    str_pos_temp = str_pos_eol + i

                                    char_height = self.text_heights[str_pos_temp]
                                    char_family = self.text_families[str_pos_temp]

                                    try:
                                        _, x_adv, scale_factor = self.get_font_char(
                                            char_family, character
                                        )
                                    except:
                                        x_adv = 0
                                        scale_factor = 1

                                    w_temp += x_adv * scale_factor * char_height

                                    i += 1
                                    if character.isspace() and not character == nbsp:
                                        break  # Break at space, except non-breaking

                                render_line = False
                                if (
                                    w_temp > rect_width
                                ):  # If the word will overflow the box
                                    if word_start == line_start:
                                        # This is the first word in the strip, so this
                                        # word(alone) is wider than the box. Render it.
                                        render_line = True
                                    else:  # Not the first word in the strip.
                                        # Render the line up UNTIL this word.
                                        render_line = True
                                        i = word_start
                                elif i >= line_length:
                                    # Render at end of text_line, if not overflowing.
                                    render_line = True

                                if render_line:
                                    # Create group for rendering a strip of text:
                                    line_group = group.add(Group())

                                    w_temp = 0
                                    w = 0

                                    self.new_line = True
                                    width_this_line = 0
                                    line_max_v_spacing = 0

                                    j = line_start

                                    while j < i:  # Calculate max height for the strip:
                                        str_pos_temp = str_pos_eol + j
                                        char_height = float(
                                            self.text_heights[str_pos_temp]
                                        )
                                        charline_spacing = float(
                                            self.text_spacings[str_pos_temp]
                                        )
                                        char_v_spacing = charline_spacing * char_height
                                        if char_v_spacing > line_max_v_spacing:
                                            line_max_v_spacing = char_v_spacing
                                        j = j + 1

                                    v = v + line_max_v_spacing

                                    char_data = dict()
                                    char_data["vertoffset"] = v
                                    char_data["stroke_scale"] = scale_r

                                    j = line_start
                                    while j < i:  # Render the strip on the page
                                        str_pos = str_pos_eol + j

                                        char_height = self.text_heights[str_pos]
                                        char_family = self.text_families[str_pos]
                                        text_align = self.text_aligns[str_pos]

                                        char_data["char"] = text_line[j]
                                        char_data["font_height"] = char_height
                                        char_data["font_family"] = char_family
                                        char_data["offset"] = w

                                        w = self.draw_svg_text(char_data, line_group)

                                        width_this_line = w

                                        j = j + 1
                                        str_pos = str_pos + 1

                                    line_start = i

                                    # Alignment for the strip:

                                    the_transform = None
                                    if text_align == "center":  # when using text-align
                                        the_transform = Transform(
                                            translate=(
                                                (float(rect_width) - width_this_line)
                                                / 2
                                            )
                                        )
                                    elif text_align == "end":
                                        the_transform = Transform(
                                            translate=(
                                                float(rect_width) - width_this_line
                                            )
                                        )
                                    if the_transform is not None:
                                        line_group.transform = the_transform

                                    if first_line:
                                        y_offs_overall = (
                                            line_max_v_spacing / 3
                                        )  # Heuristic
                                        first_line = False

                        str_pos_eol = str_pos_eol + extd_line_length
                        str_pos = str_pos_eol

                    the_transform = Transform(
                        translate=(start_x, float(start_y) - y_offs_overall)
                    )

                else:  # If this is a text object, rather than a flowroot object:
                    """
                    CASE B: Handle regular(non-flowroot) text nodes
                    """

                    # Use text-anchor, not text-align, in text(not flowroot) elements
                    text_align = node_style("text-anchor")

                    """
                    Recursively loop through content of the text object,
                    looping through text, tspan, and other things as necessary.
                    (A recursive search since style elements may be nested.)

                    Create multiple lists: One of text content, others of the
                    style that should be applied to that content.

                    For each line, want to record the plain text, font size
                    per character, text alignment, and x, y start values
                    for that line)

                    (We may need to eventually handle additional text types and
                    tags, as when importing from other SVG sources. We should
                    try to eventually support additional formulations
                    of x, y, dx, dy, etc.
                    https://www.w3.org/TR/SVG/text.html#TSpanElement)

                    then, loop through those lists, one line at a time,
                    rendering text onto lines. If the x or y values changed,
                    assume we've started a new line.

                    Note: A text element creates a single line
                    of text; it does not create multiline text by including
                    line returns within the text itself. Multiple lines of text
                    are created with multiple text or tspan elements.
                    """

                    node_info = dict()
                    node_info["font_height"] = font_height
                    node_info["font_family"] = font_family
                    node_info["anchor"] = text_align
                    node_info["x_pos"] = start_x
                    node_info["y_pos"] = start_y
                    node_info["line_spacing"] = line_spacing

                    # Keep track of line number. Used in cases where daughter
                    # tspan elements do not have Y positions given.
                    # Reset to zero on each text element.
                    self.line_number = 0

                    self.recursively_parse_text(node, node_info)
                    # self.recursively_parse_text(node, font_height, text_align, start_x, start_y)

                    if self.text_string == "":
                        continue  # No convertable text in this SVG element.
                    if self.text_string.isspace():
                        continue  # No convertable text in this SVG element.

                    letter_vals = [q for q in self.text_string]
                    str_len = len(letter_vals)

                    # Use a group for each line. This starts the first:
                    line_group = group.add(Group())

                    i = 0
                    while i < str_len:  # Loop through the entire text of the string.
                        try:
                            x_start_line = float(self.text_x[i])  # Start new line here.
                        except ValueError:
                            self.warn_unkern = True
                            return

                        y_start_line = float(self.text_y[i])

                        while i < str_len:
                            # Inner while loop, that we will break out of,
                            # back to the outer while loop.

                            q_val = letter_vals[i]
                            charfont_height = self.text_heights[i]

                            char_data = dict()
                            char_data["char"] = q_val
                            char_data["font_family"] = self.text_families[i]

                            char_data["font_height"] = charfont_height
                            char_data["offset"] = w
                            char_data["vertoffset"] = 0
                            char_data["stroke_scale"] = scale_r

                            w = self.draw_svg_text(char_data, line_group)
                            width_this_line = w
                            w_temp = w

                            # Set the alignment if(A) this is the last character in the string
                            # or if the next piece of the string is at a different position

                            set_alignment = False
                            i_next = i + 1
                            if i_next >= str_len:  # End of the string; last character.
                                set_alignment = True
                            elif (float(self.text_x[i_next]) != x_start_line) or (
                                float(self.text_y[i_next]) != y_start_line
                            ):
                                set_alignment = True

                            if set_alignment:
                                text_align = self.text_aligns[i]
                                # Not currently supporting text alignment that changes in the span;
                                # Use the text alignment as of the last character.

                                # Left(or "start") alignment is default.
                                # if(text_align == "middle"): Center alignment
                                # if(text_align == "end"): Right alignment
                                #
                                # Strategy: Align every row (left, center, or right)
                                # as it is created.

                                x_shift = 0
                                if text_align == "middle":  # when using text-anchor
                                    x_shift = x_start_line - (width_this_line / 2)
                                elif text_align == "end":
                                    x_shift = x_start_line - width_this_line
                                else:
                                    x_shift = x_start_line

                                y_shift = y_start_line

                                the_transform = Transform(translate=(x_shift, y_shift))

                                line_group.transform = the_transform

                                line_group = group.add(
                                    Group()
                                )  # Create new group for this line

                                self.new_line = True  # Used for managing indent defects
                                w = 0
                                i += 1
                                break
                            i += 1  # Only executed when set_alignment is false.

                    the_transform = Transform()

                if len(line_group) == 0:
                    parent = line_group.getparent()
                    parent.remove(line_group)

                # End cases A & B. Apply transform to text/flowroot object:

                if transform is not None:
                    result = Transform(transform) @ the_transform
                else:
                    result = the_transform

                group.transform = result

                if not self.output_generated:
                    parent = group.getparent()
                    parent.remove(group)  # remove empty group

                # Preserve original element?
                if not self.options.preserve_text and self.output_generated:
                    self.nodes_to_delete.append(node)

    def effect(self):
        """
        Main entry point; Execute the extension's function.
        """

        # Input sanitization:
        self.options.mode = self.options.mode.strip('"')
        self.options.fontface = self.options.fontface.strip('"')
        self.options.otherfont = self.options.otherfont.strip('"')
        self.options.util_mode = self.options.util_mode.strip('"')
        self.options.sample_text = self.options.sample_text.strip('"')

        self.doc_transform = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]

        self.find_font_files()

        self.handle_viewbox()

        # Calculate "ideal" effective width of rendered strokes:
        #   Default: 1/800 of page width or height, whichever is smaller

        _rendered_stroke_scale = 1 / (self.PX_PER_INCH * 800.0)

        if self.svg_width is not None:
            if self.svg_width < self.svg_height:
                self.render_width = self.svg_width * _rendered_stroke_scale
            else:
                self.render_width = self.svg_height * _rendered_stroke_scale

        if self.options.mode == "help":
            inkex.errormsg(self.help_text)
        elif self.options.mode == "utilities":
            if self.options.util_mode == "sample":
                self.font_table()
            else:
                self.glyph_table()
        else:
            if self.options.ids:
                # Traverse selected objects
                for id_ref in self.options.ids:
                    transform = self.recursive_get_encl_transform(
                        self.svg.selection[id_ref]
                    )
                    self.recursively_traverse_svg(
                        [self.svg.selection[id_ref]], transform
                    )
            else:  # Traverse entire document
                self.recursively_traverse_svg(
                    self.document.getroot(), self.doc_transform
                )

        for element_to_remove in self.nodes_to_delete:
            if element_to_remove is not None:
                parent = element_to_remove.getparent()
                if parent is not None:
                    parent.remove(element_to_remove)

        if self.font_load_fail:
            inkex.errormsg(_("Warning: unable to load SVG stroke fonts."))

        if self.warn_unkern:
            inkex.errormsg(
                _(
                    "Warning: unable to render text.\n"
                    + "Please use Text > Remove Manual Kerns to remove kerning prior to using this extension."
                )
            )

        if self.warn_unflow:
            inkex.errormsg(
                _(
                    "Warning: unable to convert text flowed into a frame.\n"
                    + "Please use Text > Unflow to convert it prior to use.\n"
                    + "If you are unable to identify the object in question, "
                    + "please contact technical support for help."
                )
            )


if __name__ == "__main__":
    Hershey().run()



# SPDX-FileCopyrightText: 2024 Software Freedom Conservancy <info@sfconservancy.org>
# SPDX-FileCopyrightText: 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Script to extract afdesign data"""

import os
import sys

import inkex


HERE = os.path.dirname(__file__) or "."
# This is suggested by https://docs.python-guide.org/writing/structure/.
sys.path.insert(0, os.path.abspath(os.path.join(HERE, "..")))

from inkaf.utils import to_pretty_xml  # noqa: E402
from inkaf.parser.extract import AFExtractor  # noqa: E402
from inkaf.svg.convert import AFConverter  # noqa: E402


class AFInput(inkex.InputExtension):
    """Input extension to extract afdesign data"""

    def add_arguments(self, pars):
        """Add command line arguments"""
        # optional arguments, see https://stackoverflow.com/a/15008806/1320237
        pars.add_argument(
            "--pretty",
            action="store_true",
            dest="pretty_print",
            default=False,
            help="Create an SVG file that has several lines and looks pretty to read.",
        )

    def load(self, stream) -> str:
        """Load the afdesign file"""
        converter = AFConverter()
        converter.convert(AFExtractor(stream))
        return self.svg_to_string(converter.doc.getroot())

    def svg_to_string(self, svg: inkex.SvgDocumentElement) -> str:
        """Convert the SvgDocumentElement to a string.

        This is mostly copied from inkex.elements._svg.SvgDocumentElement.tostring().
        """
        result = svg.tostring()
        if self.options.pretty_print:
            return to_pretty_xml(result).decode("utf-8")
        return result


def main():
    """Main entry point for the command line."""
    stdout = sys.stdout.buffer  # type: ignore
    # redirect print statements to stderr
    # see https://stackoverflow.com/a/15860430/1320237
    sys.stdout = sys.stderr
    AFInput().run(output=stdout)
    stdout.write(b"\n")
